//  Copyright (c) 2006, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"
#include "geoshader.h"
#include "p_utils.h"

#define EPS	 1e-4
#define BLACK(C) ((C).r==0 && (C).g==0 && (C).b==0)

typedef struct
{
	miColor         color;				// color of light source
	miScalar		intensity;
	miBoolean		emit_diffuse;
	miBoolean		emit_specular;
	miBoolean       shadow;				// light casts shadows
	miScalar        transparency;       // makes opaque objects transparent 
	miScalar		light_radius;		// shadow rays spread
	int				shadow_rays;		// shadow samples
	miTag			light_category;
}p_DirectTK_t;


typedef struct		//struct that p_LightTK return
{
	miColor l_color_shd;		//light color with shadow
	miColor	l_color;			//light color
	miBoolean emit_diffuse;
	miBoolean emit_specular;
	char category[32];
}light_color;

extern "C"
{
DLLEXPORT int p_DirectTK_version(void) {return(2);}

DLLEXPORT miBoolean p_DirectTK(register miColor *result, register miState *state, register p_DirectTK_t *paras)
{
	register miScalar       transp;
	miScalar				intens;
	miScalar		light_radius;
	miColor			filter;
	miColor			sum_result;
	miColor			cur_result;
	miVector		shiftVL;
	double			sample[2];
	int				sample_count = 0;
	miUint			shadow_rays;
	int TK = 1;
	light_color lightTK;
	intens = *mi_eval_scalar(&paras->intensity);
	*result = *mi_eval_color(&paras->color);
	result->r *= intens;
	result->g *= intens;
	result->b *= intens;
	if (state->type != miRAY_LIGHT)                 /* visible area light*/
		return(miTRUE);

	miColor temp_light_color;
	temp_light_color = *result;

	transp = *mi_eval_scalar(&paras->transparency);
	if (*mi_eval_boolean(&paras->shadow) && transp < 1)  // shadows
	{
		filter.r = filter.g = filter.b = filter.a = 1;
		light_radius = *mi_eval_scalar(&paras->light_radius);
		light_radius *= 0.016f;		//scale value ~ to maya's value

		cur_result = *result;
		sum_result.r = sum_result.g = sum_result.b = sum_result.a = 0;

		miVector light_dir;
		miVector original_ldir = state->dir;		//save original direction to light

		if(light_radius > 0)
		{
			//decrease shadow rays by reflect/refract depth
			shadow_rays = *mi_eval_integer(&paras->shadow_rays);
			if(shadow_rays > 1)
			{
				if(final_gather_mode(state))
					shadow_rays = 1;
				else
				{
					int level = state->refraction_level + state->reflection_level;
					shadow_rays >>= level;
					if(shadow_rays < 1)
						shadow_rays = 1;
				}
			}

			double inv_transp = 1 - transp;
			while(mi_sample(sample, &sample_count, state, 2, &shadow_rays)) 
			{
				shiftVL = p_sphere_sample(sample);
				mi_vector_mul(&shiftVL, light_radius);

				mi_vector_add(&light_dir, &state->dir, &shiftVL);		//mix random vector and light direction
				mi_vector_normalize(&light_dir);

				filter.r = filter.g = filter.b = filter.a = 1;
				state->dir = light_dir;

				if(!mi_trace_shadow(&filter, state) || BLACK(filter))	//black shadow
				{
					result->r *= transp;
					result->g *= transp;
					result->b *= transp;
				}
				else // transparent
				{
					result->r *= (float)(transp + inv_transp * filter.r);
					result->g *= (float)(transp + inv_transp * filter.g);
					result->b *= (float)(transp + inv_transp * filter.b);
				}
				sum_result.r += result->r / shadow_rays;
				sum_result.g += result->g / shadow_rays;
				sum_result.b += result->b / shadow_rays;
				*result = cur_result;
				state->dir = original_ldir;
			}
			*result = sum_result;
		}
		else
		{
			if(!mi_trace_shadow(&filter, state) || BLACK(filter))
			{
				result->r *= transp; 
				result->g *= transp;
				result->b *= transp;
				if (transp == 0 && !TK)
					return(miFALSE);
			}
			else // transparent
			{
				double inv_transp = 1 - transp;
				result->r *= (float)(transp + inv_transp * filter.r);
				result->g *= (float)(transp + inv_transp * filter.g);
				result->b *= (float)(transp + inv_transp * filter.b);
			}
		}
		state->dir = original_ldir;		//restore original light direction
	}
	if(TK)
	{
		lightTK.emit_diffuse = *mi_eval_boolean(&paras->emit_diffuse);
		lightTK.emit_specular = *mi_eval_boolean(&paras->emit_specular);
		lightTK.l_color = temp_light_color;
		lightTK.l_color_shd = *result;

		char *category = lightTK.category;
		char *eval_categ;
		miTag tag;
		if(paras->light_category)
		{
			tag = *mi_eval_tag(&paras->light_category);
			eval_categ = (char *)mi_db_access(tag);
			mi_db_unpin(tag);
			if(!eval_categ || !eval_categ[0])
			{
//				mi_warning("p_PointTK: empty light category specified.");
				for(int i = 0; i < 32; i++)
					category[i] = 0;
			}
			else
			{
				int counter = 0;
				while(*eval_categ || (counter < 31))
				{
					*category = *eval_categ;
					eval_categ++;
					category++;
					counter++;
				}
				category = NULL;
			}
		}
		else
			mi_warning("p_DirectTK: no light category specified.");

		mi_shaderstate_set(state, lightsTK_key, &lightTK, sizeof(lightTK), miSS_LIFETIME_EYERAY);
	}
	return(miTRUE);
}
}
