//  Copyright (c) 2008, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"
#include "mi_shader_if.h"
#include "p_utils.h"

#define miEPS 0.0001

#ifdef DEBUG
#define miASSERT assert
#else
#define miASSERT(x)
#endif

typedef struct
{
	miBoolean		trans_en;
	miColor			transparency_color;
	miScalar		trans_root;
	miScalar		trans_tip;
	miScalar		trans_cross;
	miScalar		trans_range;
	miScalar		trans_border;
	miInteger		trans_multiple;
	miScalar		trans_limit;
	miInteger		trans_depth;

	miColor			shadow_col;
	miScalar		shadow_scale;
	miScalar		shadow_trans;

	int		mode;			// light mode: 0..2
	int		i_light;		// index of first light
	int		n_light;		// number of lights
	miTag		light[1];	// list of lights
}p_HairTK_shadow_t;


typedef struct
{
	miScalar	contrib;
	miInteger	level;
} p_hair_info;



extern "C"
{
DLLEXPORT int p_HairTK_shadow_version(void){return 1;}

DLLEXPORT miBoolean p_HairTK_shadow(miColor *result, miState *state, p_HairTK_shadow_t *paras)
{
	miBoolean	trans_en;
	miColor		trans_col;
	miScalar	trans_root, trans_tip, trans_cross, trans_range, trans_border;
	miScalar	trans_limit;
	miInteger	trans_depth, trans_multiple;

	miScalar	trans = 0.0f;

	int			n, i, mode;
	miTag			*light;


    // illegal call, return to prevent crash
	if (state->type != miRAY_SHADOW)
		return(miFALSE);

	mode = *mi_eval_integer(&paras->mode);
	if (mode < 3) 
	{
		n = *mi_eval_integer(&paras->n_light);
		light = mi_eval_tag(paras->light) +
					*mi_eval_integer(&paras->i_light);
		switch(mode) 
		{
			case 1: mi_inclusive_lightlist(&n, &light, state); break;
			case 2: mi_exclusive_lightlist(&n, &light, state); break;
		}
		for (i=0; i < n; i++, light++)
			if (state->light_instance == *light)
				break;
		if (i == n)
			return(miTRUE);
	}

	// first of all, grab the transparency settings
	trans_en	= *mi_eval_boolean(&paras->trans_en);
	trans_limit = *mi_eval_scalar(&paras->trans_limit);
	trans_depth = *mi_eval_integer(&paras->trans_depth);

	// compute our multiplicity factor, and the crosslocation / costheta
	trans_multiple = *mi_eval_integer(&paras->trans_multiple);

	float crossloc = (float)fmod((double)(state->bary[0] * (float)trans_multiple), 1.0);
	float costheta = (float)cos(crossloc * M_PI - M_PI * 0.5);

	if (trans_en)
	{
		// evaluate all the transparency options
		trans_col	= *mi_eval_color(&paras->transparency_color);
		trans_root	= *mi_eval_scalar(&paras->trans_root);
		trans_tip	= *mi_eval_scalar(&paras->trans_tip);

		trans_range	= *mi_eval_scalar(&paras->trans_range);
		trans_cross	= *mi_eval_scalar(&paras->trans_cross);
		trans_border= *mi_eval_scalar(&paras->trans_border);

		float interp = p_smooth_step(p_clamp_0_1(trans_cross - trans_range), p_clamp_0_1(trans_cross + trans_range), state->bary[1]);

		trans = trans_root * (1.0f - interp) + trans_tip * interp;

		if (trans_border > miAlphaEPS)
			trans += (1.0f - costheta) * trans_border * (1.0f - trans);

		trans_col.r = p_max(trans_col.r, trans);
		trans_col.g = p_max(trans_col.g, trans);
		trans_col.b = p_max(trans_col.b, trans);
		trans_col.a = p_max(trans_col.a, trans);
		
		// now that we have our transparency, compute the contribution to the final pixel
//		if (state->type != miRAY_SHADOW)
//		{
//			mh_data.contrib *= p_color_maxcomp(&trans_col);
//			mh_data.level++;
//			mi_shaderstate_set(state, "p_hair_stat", &mh_data, sizeof(mh_data), miSS_LIFETIME_EYERAY);
//		}
	}

	// now that all the transparency stuff has been decided, determine if we are called as shadow shader
	miColor		shadow_col;
	miScalar	shadow_scale, shadow_trans;
	if (state->options->shadow == 's')
	{
		p_choose_volume(state);
		mi_trace_shadow_seg(result, state);
	}

	// pretty ugly but straightforward
	shadow_col = *mi_eval_color(&paras->shadow_col);
	shadow_scale = *mi_eval_scalar(&paras->shadow_scale);
	shadow_trans = *mi_eval_scalar(&paras->shadow_trans);

	shadow_col.r = shadow_col.r * shadow_scale + (1.0f - shadow_scale);
	shadow_col.g = shadow_col.g * shadow_scale + (1.0f - shadow_scale);
	shadow_col.b = shadow_col.b * shadow_scale + (1.0f - shadow_scale);

	if (trans_en)
	{
		shadow_col.r += trans_col.r * (1.0f - shadow_col.r) * shadow_trans;
		shadow_col.g += trans_col.g * (1.0f - shadow_col.g) * shadow_trans;
		shadow_col.b += trans_col.b * (1.0f - shadow_col.b) * shadow_trans;
	}

	result->r *= shadow_col.r;
	result->g *= shadow_col.g;
	result->b *= shadow_col.b;

	return(miBoolean)(result->r || result->g || result->b);
}
}
