//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifdef HPUX
#pragma OPT_LEVEL 1	/* workaround for HP/UX optimizer bug, +O2 and +O3 */
#endif

//#include <stdio.h>
//#include <stdlib.h>		/* for abs */
#include <float.h>		/* for FLT_MAX */
#include <math.h>
#include <string.h>
#include <assert.h>
#include "shader.h"
#include "p_static_param.h"
#include "p_reflect_refract.h"
#include "p_deriv.h"
#include "p_fur.h"
#include "p_occlusion.h"


typedef struct		//struct for all output channels
{
	miColor	out_result;
	miColor	out_shading;
	miColor	out_color;
	miColor	out_color_occlusion;
	miColor	out_diffuse;
	miColor	out_diffuse_shadow;
	miColor	out_specular;
	miColor	out_specular_shadow;
	miColor	out_indirect_illum;
	miColor	out_ambient;
	miColor	out_amb_occlusion;
	miColor	out_occlusion;
	miColor out_reflect;
	miColor out_refract;
	miColor out_incandescence;
	miColor out_z;
	miColor out_bent_normal;
	miColor out_specialC1;
	miColor out_specialC2;
	miColor out_specialC3;
	miColor out_specialC4;
	miColor out_specialC5;
	miColor out_specialC6;
	miColor out_specialC7;
	miColor out_specialC8;
	miColor out_specialC9;
	miColor out_specialC10;
}special_str_mega;

typedef struct		//struct that p_LightTK return
{
	miColor l_color_shd;		//light color with shadow
	miColor	l_color;			//light color
	miBoolean emit_diffuse;
	miBoolean emit_specular;
	char category[32];
}light_color;

int check_nan( float val )
{
	union {
		float         f;
		unsigned int   i;
	} number;

	number.f = val;

	// If the exponent is == 255, then we have either a
	// +/- infinite or a not-a-number. In either case,
	// return zero.
	return( (number.i & 0x7F800000UL) != 0x7F800000UL );
} 


static void choose_volume(miState *state)
{
	register int		n = 0;				// same-material counter
	register miState	*s, *s_in = 0;		// for finding enclosing mtl

	for (n=0, s=state; s; s=s->parent)		// history?
		if ((s->type == miRAY_TRANSPARENT	   ||
		     s->type == miRAY_REFRACT		   ||
		     s->type == miPHOTON_TRANSMIT_SPECULAR ||
		     s->type == miPHOTON_TRANSMIT_GLOSSY   ||
			 s->type == miPHOTON_TRANSPARENT	   ||
		     s->type == miPHOTON_TRANSMIT_DIFFUSE) &&
		     s->parent && s->parent->shader == state->shader)
		{
			n++;
			if (!s_in)
				s_in = s->parent;
		}

	if (n & 1)					// odd: exiting
		state->refraction_volume = s_in ? s_in->volume : state->camera->volume;
	else						// even: entering
		if (!state->refraction_volume)
			state->refraction_volume = state->volume;
}

inline
miVector UVblend(miVector U, miVector V, miScalar ang)
{
	miVector result;

	miScalar value = ang / 90;
	if(ang <= 90 && ang >= 0)
	{
		result.x = (U.x * (1 - value) + V.x * value);
		result.y = (U.y * (1 - value) + V.y * value);
		result.z = (U.z * (1 - value) + V.z * value);
	}
	else if(ang <= 180 && ang > 90)
	{
		value -= 1;
		mi_vector_neg(&U);
//		mi_warning("%s", ">90");
		result.x = (U.x * value + V.x * (1 - value));
		result.y = (U.y * value + V.y * (1 - value));
		result.z = (U.z * value + V.z * (1 - value));
	}
	else if(ang <= 270 && ang > 180)
	{
		value -= 2;
		mi_vector_neg(&U);
		mi_vector_neg(&V);
//		mi_warning("%s", ">180");
		result.x = (U.x * (1 - value) + V.x * value);
		result.y = (U.y * (1 - value) + V.y * value);
		result.z = (U.z * (1 - value) + V.z * value);
	}
	else if(ang <= 360 && ang > 270)
	{
		value -= 3;
		mi_vector_neg(&V);
//		mi_warning("%s", ">270");
		result.x = (U.x * value + V.x * (1 - value));
		result.y = (U.y * value + V.y * (1 - value));
		result.z = (U.z * value + V.z * (1 - value));
	}
	else
	{
		result.x = (U.x * (1 - value) + V.x * value);
		result.y = (U.y * (1 - value) + V.y * value);
		result.z = (U.z * (1 - value) + V.z * value);
	}

	return result;
}

extern "C"
{
DLLEXPORT miBoolean p_MegaTK_init(miState *state, p_MegaTK_t *paras, miBoolean *inst_req)
{
	struct static_param *sp, **spp;
	miScalar ior, ior_out;

	if (!paras)
	{
		*inst_req = miTRUE;
		return(miTRUE);
	}

	sp = (struct static_param *)
	mi_mem_allocate(sizeof(struct static_param));

	mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
	*spp = sp;

	ior	= *mi_eval_scalar(&paras->ior_in_refr);
	ior_out	= *mi_eval_scalar(&paras->ior_out_refr);

	if (ior <= 0)
	{
		mi_warning("p_MegaTK cannot use zero or negative index of refraction %f (using 1.0)", ior);
		sp->ior = 1;
	}
	else
		sp->ior = ior;

	if (ior_out < 0)
	{
		mi_warning("cannot use negative index of refraction %f for p_MegaTK (ignoring)", ior_out);
		sp->is_interface = 0;
		sp->ior_out	 = 1;
	}
	else if (ior_out == 0)
	{
		sp->is_interface = 0;
		sp->ior_out	 = 1;
	}
	else
	{
		sp->is_interface = 1;
		sp->ior_out	 = ior_out;
	}

//	sp->ignore_normals = miFALSE;

//	mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
//	*spp = sp;

	if (state->options->max_samples >= 1)
	{
		sp->russian_level = *mi_eval_integer(&paras->russian_roulette_level) - state->options->max_samples;
		if (sp->russian_level <= 1)
			sp->russian_level = 1;
	}
	else
		sp->russian_level = 4;

	sp->opacity_conversion = *mi_eval_integer(&paras->opacity_conversion);
	sp->use_constant = *mi_eval_boolean(&paras->use_constant);
	sp->use_simple_diffuse = *mi_eval_boolean(&paras->use_simple_diffuse);
	if(paras->sd_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->sd_light_category);
		sp->sd_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->use_fur_diffuse = *mi_eval_boolean(&paras->use_fur_diffuse);
	if(paras->fd_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->fd_light_category);
		sp->fd_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->use_phong_specular = *mi_eval_boolean(&paras->use_phong_specular);
	if(paras->p_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->p_light_category);
		sp->p_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->use_blinn_specular = *mi_eval_boolean(&paras->use_blinn_specular);
	if(paras->b_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->b_light_category);
		sp->b_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->use_velvet_specular = *mi_eval_boolean(&paras->use_velvet_specular);
	if(paras->v_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->v_light_category);
		sp->v_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->use_wet_specular = *mi_eval_boolean(&paras->use_wet_specular);
	if(paras->we_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->we_light_category);
		sp->we_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->use_ward_specular = *mi_eval_boolean(&paras->use_ward_specular);
	if(paras->w_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->w_light_category);
		sp->w_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->w_use_deriv = *mi_eval_boolean(&paras->w_use_deriv);
	sp->use_fur_specular = *mi_eval_boolean(&paras->use_fur_specular);
	if(paras->f_light_category)
	{
		miTag tag = *mi_eval_tag(&paras->f_light_category);
		sp->f_light_category = (char *)mi_db_access(tag);
		mi_db_unpin(tag);
	}
	sp->use_indirect_illum = *mi_eval_boolean(&paras->use_indirect_illum);
	sp->path_indirect = *mi_eval_boolean(&paras->path_indirect);
	sp->path_indirect_samples = *mi_eval_integer(&paras->path_indirect_samples);\
	sp->path_indirect_depth = *mi_eval_integer(&paras->path_indirect_depth);
	sp->ind_ignore_bump = *mi_eval_boolean(&paras->ind_ignore_bump);
	sp->use_ambient = *mi_eval_boolean(&paras->use_ambient);
	sp->use_occlusion = *mi_eval_boolean(&paras->use_occlusion);
	sp->o_trace_environment = *mi_eval_boolean(&paras->o_trace_environment);
	sp->o_inv_sec_illum = *mi_eval_boolean(&paras->o_inv_sec_illum);
	sp->o_inv_refl = *mi_eval_boolean(&paras->o_inv_refl);
	sp->o_inv_refr = *mi_eval_boolean(&paras->o_inv_refr);
	sp->o_spread = *mi_eval_scalar(&paras->o_spread);
	sp->o_samples = *mi_eval_integer(&paras->o_samples);
	sp->o_near_clip = *mi_eval_scalar(&paras->o_near_clip);
	sp->o_far_clip = *mi_eval_scalar(&paras->o_far_clip);
	sp->o_falloff_gain = *mi_eval_scalar(&paras->o_falloff_gain);
	sp->o_gamma = 1/(*mi_eval_scalar(&paras->o_gamma));
	sp->o_opacity_mode = *mi_eval_integer(&paras->o_opacity_mode);
	sp->o_bent_normal = *mi_eval_boolean(&paras->o_bent_normal);
	sp->o_bent_space = *mi_eval_integer(&paras->o_bent_space);
	sp->use_reflection = *mi_eval_boolean(&paras->use_reflection);
	sp->anisotropic_reflect = *mi_eval_boolean(&paras->anisotropic_reflect);
	sp->use_refraction = *mi_eval_boolean(&paras->use_refraction);
	sp->anisotropic_refract = *mi_eval_boolean(&paras->anisotropic_refract);
	sp->russian_roulette = *mi_eval_boolean(&paras->russian_roulette);
	sp->samples_decrease = *mi_eval_integer(&paras->samples_decrease_factor);
	sp->ignore_normals = *mi_eval_boolean(&paras->ignore_normals);
	sp->use_incandescence = *mi_eval_boolean(&paras->use_incandescence);
	sp->use_z = *mi_eval_boolean(&paras->use_z);
	sp->opacity_z = *mi_eval_boolean(&paras->opacity_z);
	sp->reflection_z = *mi_eval_boolean(&paras->reflection_z);
	sp->refraction_z = *mi_eval_boolean(&paras->refraction_z);
	sp->use_float_z = *mi_eval_boolean(&paras->use_float_z);
	sp->invert_z = *mi_eval_boolean(&paras->invert_z);

	sp->use_SpecialC1 = *mi_eval_boolean(&paras->use_SpecialC1);
	sp->use_SpecialC2 = *mi_eval_boolean(&paras->use_SpecialC2);
	sp->use_SpecialC3 = *mi_eval_boolean(&paras->use_SpecialC3);
	sp->use_SpecialC4 = *mi_eval_boolean(&paras->use_SpecialC4);
	sp->use_SpecialC5 = *mi_eval_boolean(&paras->use_SpecialC5);
	sp->use_SpecialC6 = *mi_eval_boolean(&paras->use_SpecialC6);
	sp->use_SpecialC7 = *mi_eval_boolean(&paras->use_SpecialC7);
	sp->use_SpecialC8 = *mi_eval_boolean(&paras->use_SpecialC8);
	sp->use_SpecialC9 = *mi_eval_boolean(&paras->use_SpecialC9);
	sp->use_SpecialC10 = *mi_eval_boolean(&paras->use_SpecialC10);

	sp->use_normal = *mi_eval_boolean(&paras->use_normal);
	sp->add_normal = *mi_eval_boolean(&paras->add_to_normal);


	size_t index = -1;
	sp->shading_pass = *mi_eval_boolean(&paras->shading_pass);
	sp->shading_buf = (int)index;
	sp->color_pass = *mi_eval_boolean(&paras->color_occlusion_pass);
	sp->color_buf = (int)index;
	sp->diffuse_pass = *mi_eval_boolean(&paras->diffuse_pass);
	sp->diffuse_buf = (int)index;	
	sp->diffuse_shadow_pass = *mi_eval_boolean(&paras->diffuse_shadow_pass);
	sp->diffuse_shadow_buf  = (int)index;	
	sp->specular_pass = *mi_eval_boolean(&paras->specular_pass);
	sp->specular_buf = (int)index;
	sp->specular_shadow_pass = *mi_eval_boolean(&paras->specular_shadow_pass);
	sp->specular_shadow_buf = (int)index;
	sp->occlusion_pass = *mi_eval_boolean(&paras->occlusion_pass);
	sp->occlusion_buf = (int)index;
	sp->indirect_illum_pass = *mi_eval_boolean(&paras->indirect_illum_pass);
	sp->indirect_illum_buf = (int)index;
	sp->ambient_pass = *mi_eval_boolean(&paras->ambient_pass);
	sp->ambient_buf = (int)index;
	sp->amb_occlusion_pass = *mi_eval_boolean(&paras->amb_occlusion_pass);
	sp->amb_occlusion_buf = (int)index;
	sp->reflect_pass = *mi_eval_boolean(&paras->reflect_pass);
	sp->reflect_buf = (int)index;
	sp->refract_pass = *mi_eval_boolean(&paras->refract_pass);
	sp->refract_buf = (int)index;
	sp->incandescence_pass = *mi_eval_boolean(&paras->incandescence_pass);
	sp->incandescence_buf = (int)index;
	sp->z_pass = *mi_eval_boolean(&paras->z_pass);
	sp->z_buf = (int)index;
	sp->bent_normal_pass = *mi_eval_boolean(&paras->bent_normal_pass);
	sp->bent_normal_buf = (int)index;
	sp->specialC1_pass = *mi_eval_boolean(&paras->specialC1_pass);
	sp->specialC1_buf = (int)index;
	sp->specialC2_pass = *mi_eval_boolean(&paras->specialC2_pass);
	sp->specialC2_buf = (int)index;
	sp->specialC3_pass = *mi_eval_boolean(&paras->specialC3_pass);
	sp->specialC3_buf = (int)index;
	sp->specialC4_pass = *mi_eval_boolean(&paras->specialC4_pass);
	sp->specialC4_buf = (int)index;
	sp->specialC5_pass = *mi_eval_boolean(&paras->specialC5_pass);
	sp->specialC5_buf = (int)index;
	sp->specialC6_pass = *mi_eval_boolean(&paras->specialC6_pass);
	sp->specialC6_buf = (int)index;
	sp->specialC7_pass = *mi_eval_boolean(&paras->specialC7_pass);
	sp->specialC7_buf = (int)index;
	sp->specialC8_pass = *mi_eval_boolean(&paras->specialC8_pass);
	sp->specialC8_buf = (int)index;
	sp->specialC9_pass = *mi_eval_boolean(&paras->specialC9_pass);
	sp->specialC9_buf = (int)index;
	sp->specialC10_pass = *mi_eval_boolean(&paras->specialC10_pass);
	sp->specialC10_buf = (int)index;

/*
	mi::shader::Access_fb fbc(state->camera->buffertag);
	size_t index = -1;

	sp->shading_pass = *mi_eval_boolean(&paras->shading_pass);
	fbc->get_index("tk_shading", index); 
	sp->shading_buf = (int)index;

	sp->color_pass = *mi_eval_boolean(&paras->color_occlusion_pass);
	index = -1;
	fbc->get_index("tk_color", index); 
	sp->color_buf = (int)index;

	sp->diffuse_pass = *mi_eval_boolean(&paras->diffuse_pass);
	index = -1;
	fbc->get_index("tk_diffuse", index); 
	sp->diffuse_buf = (int)index;
	
	sp->diffuse_shadow_pass = *mi_eval_boolean(&paras->diffuse_shadow_pass);
	index = -1;
	fbc->get_index("tk_shadow", index); 
	sp->diffuse_shadow_buf  = (int)index;
	
	sp->specular_pass = *mi_eval_boolean(&paras->specular_pass);
	index = -1;
	fbc->get_index("tk_specular", index); 
	sp->specular_buf = (int)index;

	sp->specular_shadow_pass = *mi_eval_boolean(&paras->specular_shadow_pass);
	index = -1;
	fbc->get_index("tk_spec_shd", index); 
	sp->specular_shadow_buf = (int)index;

	sp->occlusion_pass = *mi_eval_boolean(&paras->occlusion_pass);
	index = -1;
	fbc->get_index("tk_diff_occl", index); 
	sp->occlusion_buf = (int)index;

	sp->indirect_illum_pass = *mi_eval_boolean(&paras->indirect_illum_pass);
	index = -1;
	fbc->get_index("tk_indirect", index); 
	sp->indirect_illum_buf = (int)index;

	sp->ambient_pass = *mi_eval_boolean(&paras->ambient_pass);
	index = -1;
	fbc->get_index("tk_ambient", index); 
	sp->ambient_buf = (int)index;

	sp->amb_occlusion_pass = *mi_eval_boolean(&paras->amb_occlusion_pass);
	index = -1;
	fbc->get_index("tk_amb_occlus", index); 
	sp->amb_occlusion_buf = (int)index;

	sp->reflect_pass = *mi_eval_boolean(&paras->reflect_pass);
	index = -1;
	fbc->get_index("tk_reflect", index); 
	sp->reflect_buf = (int)index;

	sp->refract_pass = *mi_eval_boolean(&paras->refract_pass);
	index = -1;
	fbc->get_index("tk_refract", index); 
	sp->refract_buf = (int)index;

	sp->incandescence_pass = *mi_eval_boolean(&paras->incandescence_pass);
	index = -1;
	fbc->get_index("tk_incand", index); 
	sp->incandescence_buf = (int)index;

	sp->z_pass = *mi_eval_boolean(&paras->z_pass);
	index = -1;
	fbc->get_index("tk_z", index); 
	sp->z_buf = (int)index;

	sp->bent_normal_pass = *mi_eval_boolean(&paras->bent_normal_pass);
	index = -1;
	fbc->get_index("tk_bent", index); 
	sp->bent_normal_buf = (int)index;

	sp->specialC1_pass = *mi_eval_boolean(&paras->specialC1_pass);
	index = -1;
	fbc->get_index("tk_specialC1", index); 
	sp->specialC1_buf = (int)index;

	sp->specialC2_pass = *mi_eval_boolean(&paras->specialC2_pass);
	index = -1;
	fbc->get_index("tk_specialC2", index); 
	sp->specialC2_buf = (int)index;

	sp->specialC3_pass = *mi_eval_boolean(&paras->specialC3_pass);
	index = -1;
	fbc->get_index("tk_specialC3", index); 
	sp->specialC3_buf = (int)index;

	sp->specialC4_pass = *mi_eval_boolean(&paras->specialC4_pass);
	index = -1;
	fbc->get_index("tk_specialC4", index); 
	sp->specialC4_buf = (int)index;

	sp->specialC5_pass = *mi_eval_boolean(&paras->specialC5_pass);
	index = -1;
	fbc->get_index("tk_specialC5", index); 
	sp->specialC5_buf = (int)index;

	sp->specialC6_pass = *mi_eval_boolean(&paras->specialC6_pass);
	index = -1;
	fbc->get_index("tk_specialC6", index); 
	sp->specialC6_buf = (int)index;

	sp->specialC7_pass = *mi_eval_boolean(&paras->specialC7_pass);
	index = -1;
	fbc->get_index("tk_specialC7", index); 
	sp->specialC7_buf = (int)index;

	sp->specialC8_pass = *mi_eval_boolean(&paras->specialC8_pass);
	index = -1;
	fbc->get_index("tk_specialC8", index); 
	sp->specialC8_buf = (int)index;

	sp->specialC9_pass = *mi_eval_boolean(&paras->specialC9_pass);
	index = -1;
	fbc->get_index("tk_specialC9", index); 
	sp->specialC9_buf = (int)index;

	sp->specialC10_pass = *mi_eval_boolean(&paras->specialC10_pass);
	index = -1;
	fbc->get_index("tk_specialC10", index); 
	sp->specialC10_buf = (int)index;
*/
	return(miTRUE);
}

DLLEXPORT miBoolean p_MegaTK_exit(miState *state, p_MegaTK_t *paras)
{
	struct static_param **spp;

	if (paras)
	{
		mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
		mi_mem_release(*spp);
		*spp = NULL;
	}
	return(miTRUE);
}

DLLEXPORT int p_MegaTK_version(void) {return(3);}

DLLEXPORT miBoolean p_MegaTK(miColor *result, miState *state, p_MegaTK_t *paras)
{
	miColor		color;		// color from irradiance
	light_color	light_col;		//struct for light shader return
	light_color *lightTK_pt;
//	int size_light_color = 0;
	miColor		sum_light;		// summed sample colors
	miColor		sum_light_shd;		// summed sample colors
	miColor		sum_specular;		// summed specular colors
	miColor		sum_specular_shd;		// summed specular colors
	miVector	dir;		// direction towards light
	miScalar	dot_nl;		// dot prod of normal and dir
	miScalar	geom_dot_nl;
	miScalar	s;		// amount of specular reflection
//	miVector	l;			// light direction 

    special_str_mega special;		//struct for output special channel

   if(state->type == miRAY_DISPLACE)
		return(miFALSE);
   
	struct static_param *sp, **spp;
	miASSERT(state->shader);
	mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
	sp = *spp;

	miVector bump_normal;
	miVector orig_normal;
	if(sp->use_normal)					//Change normal(bump)
	{
		if(state->type == miRAY_FINALGATHER && sp->ind_ignore_bump)
			orig_normal = state->normal;
		else
		{
			orig_normal = state->normal;
			bump_normal = *mi_eval_vector(&paras->input_normal);
			if(sp->add_normal)
				mi_vector_add(&bump_normal, &state->normal, &bump_normal);
			mi_vector_normalize(&bump_normal);
			state->dot_nd = mi_vector_dot(&bump_normal, &state->dir);
			state->normal = bump_normal;
		}
	}
	else
		orig_normal = state->normal;

	float alpha;														//Alpha
	miColor *opacit;
	opacit = mi_eval_color(&paras->cOpacity);
	switch(sp->opacity_conversion)
	{
		case 0:		//r
			alpha = opacit->r;
			break;
		case 1:		//g
			alpha = opacit->g;
			break;
		case 2:		//b
			alpha = opacit->b;
			break;
		case 3:		//a
			alpha = opacit->a;
			break;
		case 4:		//average
			alpha = (opacit->r + opacit->g + opacit->b)/3;
			break;
		case 5:		//luminosity
			alpha = opacit->r*0.299f + opacit->g*0.587f + opacit->b*0.114f;
			break;
		default:	//default average
			alpha = (opacit->r + opacit->g + opacit->b)/3;
			break;
	}
	float inv_alpha = 1 - alpha;
	if(alpha > 1)
		alpha = 1;
	else if(alpha < 0)
			alpha = 0;


	if(state->type == miRAY_SHADOW)											//shadow call
	{
		if(sp->use_refraction && alpha > miAlphaEPS)
		{
			miScalar coeff_refract = *mi_eval_scalar(&paras->cRefract);
			if(coeff_refract > miEPS)
			{
				miScalar fresnel = 1;
				miScalar fresnel_refr_coeff = *mi_eval_scalar(&paras->fresnel_refr);

				//	fresnel effect
				if(fresnel_refr_coeff > miEPS)
				{
					miVector V = state->dir;
					miVector N = state->normal;
					mi_vector_neg(&V);
					mi_vector_normalize(&V);
					mi_vector_normalize(&N);
					fresnel = mi_vector_dot(&V, &N);
					if(fresnel < 0)
						fresnel = -fresnel;
					fresnel = pow(fresnel, 0.4f);
					fresnel = pow(fresnel, sp->ior/sp->ior_out);
				}
				if(fresnel > 1)
					fresnel = 1;
				else if (fresnel < 0)
					fresnel = 0;
				coeff_refract *= fresnel * fresnel_refr_coeff + (1-fresnel_refr_coeff);

				if(alpha > miAlphaEPS && coeff_refract > 0.001)
				{
					coeff_refract = 1-coeff_refract;
					coeff_refract *= alpha;
					inv_alpha = 1-coeff_refract;
				}
			}
		}
		if(state->options->shadow == 's')	//segment shadow
		{
			if(inv_alpha > miAlphaEPS)
			{
				choose_volume(state);
				mi_trace_shadow_seg(result, state);
			}
		}

		result->r *= inv_alpha;
		result->g *= inv_alpha;
		result->b *= inv_alpha;
		return(miBoolean)(result->r || result->g || result->b);
	}


	//check ray type, if it's occlusion ray, just use more simple shader
	int *rayOcc;
	int size_rayOcc = 0;
	rayOcc = (int *)mi_shaderstate_get(state, "ra_o", &size_rayOcc);  //"ra_o" used for define occlusion ray
	if (rayOcc != NULL)			//yes, it's occlusion ray, compute only alpha and return
	{
		result->r = result->g = result->b = result->a = 0;
		if(alpha < 0.999)
		{
			if(mi_trace_transparent(result, state))
			{
				result->r = result->g = result->b = 0;
				result->a *= inv_alpha;
			}
		}
		result->a += alpha;
		return(miTRUE);
	}


	if(alpha > miEPS)
		special.out_color = *mi_eval_color(&paras->cSurface);
	else
		special.out_color.r = special.out_color.g = special.out_color.b = 0;



	float refl_refr = 0;										//Amount of Reflect/Refract
	miScalar coeff_reflect = 0;	
	miScalar coeff_refract = 0;
	miBoolean do_reflect = miTRUE;
	miBoolean do_refract = miTRUE;

	//	fresnel effect
	float fresnel = 0;
	float fresnel_refl_coeff = *mi_eval_scalar(&paras->fresnel_refl);
	float fresnel_refr_coeff = *mi_eval_scalar(&paras->fresnel_refr);
    miScalar ior_in = sp->ior;
    miScalar ior_out = sp->ior_out;

	miCBoolean enter_surf = miTRUE;
	float rr_scale = 1.0;
	if(sp->use_reflection || sp->use_refraction)
	{
        enter_surf = locate_volume(state, ior_in, ior_out, sp->ignore_normals);
        fresnel = mi_fresnel_reflection(state, state->ior_in, state->ior);
//		mi_warning("p_MegaTK: %f", fresnel);
		if(fresnel > 1)
			fresnel = 1;
		else if (fresnel < 0)
			fresnel = 0;

		if(sp->use_reflection)
		{
			coeff_reflect = *mi_eval_scalar(&paras->cReflect);
			if(fresnel_refl_coeff > miEPS)
				coeff_reflect *= fresnel * fresnel_refl_coeff + (1 - fresnel_refl_coeff);
		}

		if(sp->use_refraction)
		{
			coeff_refract = *mi_eval_scalar(&paras->cRefract);
			if(fresnel_refr_coeff > miEPS)
				coeff_refract *= (1.0f - fresnel) * fresnel_refr_coeff + (1.0f - fresnel_refr_coeff);
		}


		if((sp->russian_roulette == miTRUE) && (state->reflection_level + state->refraction_level >= sp->russian_level) && sp->use_reflection && sp->use_refraction && miRAY_PHOTON(state->type) == miFALSE)	//russian roullete optimization
		{
			miColor	rspec, tspec;
			miRay_type	rt;
			rspec.r = rspec.g = rspec.b = coeff_reflect;
			tspec.r = tspec.g = tspec.b = coeff_refract;
	//		mi_warning("Before: %f", tspec.r);
	//		rt = mi_choose_simple_scatter_type(state, NULL, &rspec, NULL, &tspec);
			rt = p_choose_simple_scatter_type(state, NULL, NULL, &rspec, &tspec);
	//		mi_warning("After: %f", tspec.r);
			// mi_choose_simple_scatter_type scales
			if (rt == miPHOTON_REFLECT_SPECULAR)
			{
				do_refract = miFALSE;
				if(coeff_reflect)
					rr_scale = rspec.r / coeff_reflect;
			//		rr_scale = rspec.r / fresnel;
			}
			else if (rt == miPHOTON_TRANSMIT_SPECULAR)
			{
				do_reflect = miFALSE;
				if(coeff_refract)
					rr_scale = tspec.r / coeff_refract;
			//		rr_scale = tspec.r / (1-fresnel);
			}
			else
				do_reflect = do_refract = miFALSE;
		}
		if(sp->use_reflection)
			refl_refr += coeff_reflect * (*mi_eval_scalar(&paras->diff_minus_reflect));
		if(sp->use_refraction)
			refl_refr += coeff_refract * (*mi_eval_scalar(&paras->diff_minus_refract));
	}

	miScalar coeff_dif_simp = 0;											//Simple Diffuse
	miScalar sd_roughness = 0;
	float dot_nl_rougness = 0;
	if(sp->use_simple_diffuse && alpha > miEPS)
	{
		coeff_dif_simp = *mi_eval_scalar(&paras->cds);
		coeff_dif_simp *= 1 -refl_refr;
		sd_roughness = *mi_eval_scalar(&paras->sd_roughness) + 1;
		sd_roughness = 1/sd_roughness;
	}


																			//photon call
	if (miRAY_PHOTON(state->type))
	{
		miRay_type	type;
		miColor		photon_color;
		miVector	photon_dir;
		miColor		diffuse = {0, 0, 0, 0};
//		miColor		spec_reflect_color = {0, 0, 0, 0};
//		miColor		glos_reflect_color = {0, 0, 0, 0};
//		float		coeff = 0;

		if(sp->use_simple_diffuse && alpha > miEPS && coeff_dif_simp > miEPS)					//Diffuse photon
		{

			diffuse.r = special.out_color.r * coeff_dif_simp * alpha;
			diffuse.g = special.out_color.g * coeff_dif_simp * alpha;
			diffuse.b = special.out_color.b * coeff_dif_simp * alpha;		

			if (diffuse.r > miEPS || diffuse.g > miEPS || diffuse.b > miEPS) 
			{
				miState	*parent = state;		    
				while(parent->type == miPHOTON_TRANSPARENT && parent->parent) 
				{
					miASSERT(parent->parent != parent);
					parent = parent->parent;
				}
				if (parent->type != miPHOTON_LIGHT)
					mi_store_photon(result, state);
			}
		}

//		type = mi_choose_scatter_type(state, 0, &special.out_color, &glos_reflect_color, &spec_reflect_color);
//		mi_choose_simple_scatter_type(state, NULL, &rspec, NULL, &tspec);
//		type = mi_choose_simple_scatter_type(state, &special.out_color, &spec_reflect_color, &spec_reflect_color, &spec_reflect_color);
	
		miColor reflection_color = {0, 0, 0, 0};
		if(sp->use_reflection && coeff_reflect > miEPS && alpha > miEPS)
		{
			reflection_color = *mi_eval_color(&paras->glossy_refl);
			reflection_color.r *= coeff_reflect * alpha;
			reflection_color.g *= coeff_reflect * alpha;
			reflection_color.b *= coeff_reflect * alpha;
		}

		miColor refraction_color = {0, 0, 0, 0};
		if(sp->use_refraction && coeff_refract > miEPS && alpha > miEPS)
		{
			refraction_color = *mi_eval_color(&paras->glossy_refr);
			refraction_color.r *= coeff_refract * alpha;
			refraction_color.g *= coeff_refract * alpha;
			refraction_color.b *= coeff_refract * alpha;
		}

		type = p_choose_simple_scatter_type(state, &inv_alpha, &diffuse, &reflection_color, &refraction_color);

		miColor abs_col = {0, 0, 0, 0};
		float dist = (float)state->dist;
		switch (type) 
		{
			case miPHOTON_ABSORB:		    // no reflection or transmission
				return(miTRUE);
			case miPHOTON_TRANSPARENT:
				photon_color.r = result->r * inv_alpha;
				photon_color.g = result->g * inv_alpha;
				photon_color.b = result->b * inv_alpha;
				break;
			case miPHOTON_REFLECT_DIFFUSE:
				photon_color.r = result->r * special.out_color.r;
				photon_color.g = result->g * special.out_color.g;
				photon_color.b = result->b * special.out_color.b;
				break;
			case miPHOTON_TRANSMIT_DIFFUSE:
				photon_color.r = result->r * special.out_color.r;
				photon_color.g = result->g * special.out_color.g;
				photon_color.b = result->b * special.out_color.b;
				break;
			case miPHOTON_REFLECT_GLOSSY:
			case miPHOTON_TRANSMIT_GLOSSY:
				return(miTRUE);
			case miPHOTON_REFLECT_SPECULAR:
				photon_color = *result;
				if (!enter_surf){
					abs_col = *mi_eval_color(&paras->refr_absorb_in);
					absorb(&photon_color, &abs_col, dist);
				} else  if (sp->is_interface) {
					abs_col = *mi_eval_color(&paras->refr_absorb_out);
					absorb(&photon_color, &abs_col, dist);
				}
				photon_color.r *= reflection_color.r;
				photon_color.g *= reflection_color.g;
				photon_color.b *= reflection_color.b;
				break;
			case miPHOTON_TRANSMIT_SPECULAR:
				photon_color = *result;
				if (!enter_surf){
					abs_col = *mi_eval_color(&paras->refr_absorb_in);
					absorb(&photon_color, &abs_col, dist);
				} else if (sp->is_interface){
					abs_col = *mi_eval_color(&paras->refr_absorb_out);
					absorb(&photon_color, &abs_col, dist);
				}
				photon_color.r *= refraction_color.r;
				photon_color.g *= refraction_color.g;
				photon_color.b *= refraction_color.b;
				break;
			default:			  // Unknown scatter type
				mi_error("p_MegaTK: Unknown scatter type in photon shader!");
				return(miFALSE);
		}

		// Shoot new photon in a direction determined by the scattering type 
		switch (type) 
		{
			case miPHOTON_TRANSMIT_DIFFUSE:  // diffuse transm. (translucency)
				mi_transmission_dir_diffuse(&photon_dir, state);
				return(mi_photon_transmission_diffuse(&photon_color, state, &photon_dir));
			case miPHOTON_TRANSPARENT:		//trasparency
				return(mi_photon_transparent(&photon_color, state));
			case miPHOTON_REFLECT_DIFFUSE: // diffuse reflection (Lambert)
				mi_reflection_dir_diffuse(&photon_dir, state);
				return(mi_photon_reflection_diffuse(&photon_color, state, &photon_dir));

//			case miPHOTON_TRANSMIT_GLOSSY: // glossy transmission (Ward model)
//				if (dgs_transmit_glossy_dir(state, &photon_dir, &m))
//					return(mi_photon_transmission_glossy(&photon_color, state, &photon_dir));
			// else: full reflection case, no break ! 
//			case miPHOTON_REFLECT_GLOSSY:  // glossy reflection (Ward model)
//				dgs_reflect_glossy_dir(state, &photon_dir, &m);
//				return(mi_photon_reflection_glossy(&photon_color, state, &photon_dir));

			case miPHOTON_TRANSMIT_SPECULAR:// specular transmission
				if (mi_transmission_dir_specular(&photon_dir, state, state->ior_in, state->ior)) 
					return(mi_photon_transmission_specular(&photon_color, state, &photon_dir));

			// full reflection case, no break !
			case miPHOTON_REFLECT_SPECULAR:   // specular reflection (mirror)
				mi_reflection_dir_specular(&photon_dir, state);
				return(mi_photon_reflection_specular(&photon_color, state, &photon_dir));

			default:
				miASSERT(0); // cannot be here
				return(miFALSE);
		}
		return(miTRUE);
	}//end photon


	miScalar coeffInc = *mi_eval_scalar(&paras->cInc);													//Incandescence
	special.out_incandescence.r = special.out_incandescence.g = special.out_incandescence.b = 0;
	if(sp->use_incandescence && (coeffInc != 0) && (alpha > miEPS))
	{
		special.out_incandescence = *mi_eval_color(&paras->incandescence);
		special.out_incandescence.r *= coeffInc;
		special.out_incandescence.g *= coeffInc;
		special.out_incandescence.b *= coeffInc;
	}

	miScalar const_coeff = *mi_eval_scalar(&paras->cConst);			//Constant
	if(sp->use_constant && (const_coeff != 0) && (alpha > miEPS))
	{
		miColor *const_color;
		const_color = mi_eval_color(&paras->c_color);
		special.out_diffuse_shadow.r = special.out_diffuse.r = const_color->r * const_coeff;
		special.out_diffuse_shadow.g = special.out_diffuse.g = const_color->g * const_coeff;
		special.out_diffuse_shadow.b = special.out_diffuse.b = const_color->b * const_coeff;
	}
	else
	{
		special.out_diffuse_shadow.r = special.out_diffuse_shadow.g = special.out_diffuse_shadow.b = 0;
		special.out_diffuse.r = special.out_diffuse.g = special.out_diffuse.b = 0;
	}
	special.out_shading.r = special.out_shading.g = special.out_shading.b = 0;
	special.out_indirect_illum.r = special.out_indirect_illum.g = special.out_indirect_illum.b = 0;
	special.out_ambient.r = special.out_ambient.g = special.out_ambient.b = 0;
	special.out_amb_occlusion.r = special.out_amb_occlusion.g = special.out_amb_occlusion.b = 0;
	special.out_specular.r = special.out_specular.g = special.out_specular.b = 0;
	special.out_specular_shadow.r = special.out_specular_shadow.g = special.out_specular_shadow.b = 0;
	special.out_reflect.r = special.out_reflect.g = special.out_reflect.b = 0;
	special.out_refract.r = special.out_refract.g = special.out_refract.b = 0;
	special.out_bent_normal.r = special.out_bent_normal.g =  special.out_bent_normal.b = 0;
	special.out_z.r = special.out_z.g = special.out_z.b = 0;
	special.out_occlusion.r = special.out_occlusion.g = special.out_occlusion.b = 1;
	special.out_specialC1.r = special.out_specialC1.g = special.out_specialC1.b = 0;
	special.out_specialC2.r = special.out_specialC2.g = special.out_specialC2.b = 0;
	special.out_specialC3.r = special.out_specialC3.g = special.out_specialC3.b = 0;
	special.out_specialC4.r = special.out_specialC4.g = special.out_specialC4.b = 0;
	special.out_specialC5.r = special.out_specialC5.g = special.out_specialC5.b = 0;
	special.out_specialC6.r = special.out_specialC6.g = special.out_specialC6.b = 0;
	special.out_specialC7.r = special.out_specialC7.g = special.out_specialC7.b = 0;
	special.out_specialC8.r = special.out_specialC8.g = special.out_specialC8.b = 0;
	special.out_specialC9.r = special.out_specialC9.g = special.out_specialC9.b = 0;
	special.out_specialC10.r = special.out_specialC10.g = special.out_specialC10.b = 0;
	result->r = result->g = result->b = result->a = 0;

	int			mode;
	miTag		*light;
	int			n_light;
	int			i_light;
//	int			n;			/* light counter */
	int			samples;	/* # of samples taken */
	n_light   = *mi_eval_integer(&paras->n_light);
	i_light   = *mi_eval_integer(&paras->i_light);
	light =  mi_eval_tag(paras->light) + i_light;
	mode = *mi_eval_integer(&paras->mode);

	miVector V = state->dir;
	mi_vector_neg(&V);
	mi_vector_normalize(&V);


/*
	#ifdef MAYA
		miVector T_deriv = state->bump_x_list[0];
	#else
		miVector T_deriv;
		miVector dummy;
		BumpUV( T_deriv, dummy, state, 0);
//		T.normalize();
		mi_vector_normalize(&T_deriv);
	#endif
*/
	miVector deriv_u, deriv_v;

//	#ifdef MAX_XSI
		if(sp->use_fur_diffuse || sp->use_fur_specular)
			p_UV_orientation(&deriv_u, &deriv_v, state);
//	#else	//maya
//		Looks like some bug here... result is different in some buckets
//		deriv_u = state->bump_x_list[0];
//		deriv_v = state->bump_y_list[0];
//	#endif

	miScalar	coeff_dif_fur;											//Fur Diffuse
	miColor 	*f_skinCol;
	miScalar 	f_hairDens;
	miScalar 	f_backwardScater;
	miScalar 	f_forwardScatter;
	miScalar 	f_hairOverHairShad;
	miScalar 	f_shadowAtten;
	miScalar 	f_startShadowCo;
	miScalar 	f_endShadowCo;
	miScalar 	f_hairLeng;
	miScalar 	f_hairBaseRad;
	miScalar 	f_hairTipRad;
	miVector T;
	miScalar hair_area;
	miVector hairN = state->normal;
//	miVector skinN = state->normal;
	mi_vector_normalize(&hairN);
//	miVector TcV;
	miScalar v = state->bary[1];
	if(sp->use_fur_diffuse)
	{
		coeff_dif_fur = *mi_eval_scalar(&paras->cdf);
//		f_d_angl = *mi_eval_scalar(&paras->f_d_angle);
		f_skinCol = mi_eval_color(&paras->f_skinColor);
		f_hairDens = *mi_eval_scalar(&paras->f_hairDensity);
		f_backwardScater = *mi_eval_scalar(&paras->f_backwardScatering);
		f_forwardScatter = *mi_eval_scalar(&paras->f_forwardScattering);
		f_hairOverHairShad = *mi_eval_scalar(&paras->f_hairOverHairShadow);
		f_shadowAtten = *mi_eval_scalar(&paras->f_shadowAttenuation);
		f_startShadowCo = *mi_eval_scalar(&paras->f_startShadowCos);
		f_endShadowCo = *mi_eval_scalar(&paras->f_endShadowCos);
		f_hairLeng = *mi_eval_scalar(&paras->f_hairLength);
		f_hairBaseRad = *mi_eval_scalar(&paras->f_hairBaseRadius);
		f_hairTipRad = *mi_eval_scalar(&paras->f_hairTipRadius);
//		T = state->bump_x_list[0];
		T = UVblend(deriv_u, deriv_v, *mi_eval_scalar(&paras->f_d_angle));
		T = shiftTangent(T, hairN, 0);
		hair_area = (f_hairLeng * (f_hairBaseRad + f_hairTipRad) ) / 2.0f;
//		mi_vector_prod(&TcV, &T, &V);
	}


	miScalar	*p_coeffS;											//Phong
	miColor		*p_spec;
	miScalar	expo;				// Phong exponent (cosine power)
	p_coeffS = mi_eval_scalar(&paras->p_cS);
	p_spec = mi_eval_color(&paras->p_specular);
	expo = *mi_eval_scalar(&paras->p_exponent);



	float E;
	miVector H, Ln;
	float NH, NH2, NHSQ, Dd, Gg, VN, VH, LN, Ff, tmp;
	miScalar	*b_coeffS;												//Blinn
	miColor		*b_spec;
	miScalar	b_eccent;
	miScalar	b_rolloff;
	b_coeffS = mi_eval_scalar(&paras->b_cS);
	b_spec = mi_eval_color(&paras->b_specular);
	b_eccent = *mi_eval_scalar(&paras->b_eccentricity);
	b_rolloff = *mi_eval_scalar(&paras->b_specularRollOff);
	if(b_eccent != 1)
		E = 1 / (b_eccent * b_eccent - 1);
	else
		E = -1e5;
	VN = mi_vector_dot(&V, &state->normal);

	miScalar	*v_coeffS;												//Velvet
	miColor		*v_sheen_color;
	miScalar	*v_backscat;
	miScalar	*v_edgin;
	miScalar	*v_rough;
	v_coeffS = mi_eval_scalar(&paras->v_cS);
	v_sheen_color = mi_eval_color(&paras->v_sheen_color);
	v_backscat = mi_eval_scalar(&paras->v_backscatter);
	v_edgin = mi_eval_scalar(&paras->v_edginess);
	v_rough = mi_eval_scalar(&paras->v_roughness);

	miScalar cosine = -state->dot_nd;
	miScalar sine = sqrt( 1.0f - cosine*cosine );
	miScalar iRoughness = 1.0f / *v_rough;


	miScalar	we_cS;											//Wet
	miColor		*we_specular;
	miScalar	we_roughness;
	miScalar	we_sharpness;
	miScalar	we_falloff;
	miScalar	we_value;
//	miVector	we_refl_dir;
//	miScalar	we_rfl_dot;
	if(sp->use_wet_specular)
	{
		we_cS = *mi_eval_scalar(&paras->we_cS);
		we_specular = mi_eval_color(&paras->we_specular);
		we_roughness = 1 - *mi_eval_scalar(&paras->we_roughness)* 0.5f;
		we_roughness = pow(we_roughness, 0.3f);
		we_sharpness = (1 - *mi_eval_scalar(&paras->we_sharpness)) * 0.25f;
	}

	miScalar	w_coeffS;											//Ward
	miColor		*w_glos;
	miScalar	w_shin_u;
	miScalar	w_shin_v;
	if(sp->use_ward_specular)
	{
		w_coeffS = *mi_eval_scalar(&paras->w_cS);
		w_glos = mi_eval_color(&paras->w_specular);
		w_shin_u = *mi_eval_scalar(&paras->w_shiny_u);
		w_shin_v = *mi_eval_scalar(&paras->w_shiny_v);
	}
	float w_spec;


	miScalar	f_coeffS;											//Fur Specular
	miColor 	*f_spec1Col;
	miScalar 	f_spec1Exp;
	miColor 	*f_spec2Col;
	miScalar 	f_spec2Exp;
	miScalar 	f_tangentBreakupAm;
	miVector Ts;

	const miScalar bup = *mi_eval_scalar(&paras->f_tangentBreakup) * v;
	if(sp->use_fur_specular)
	{
		f_coeffS = *mi_eval_scalar(&paras->f_cS);
		f_spec1Col = mi_eval_color(&paras->f_specular1Color);
		f_spec1Exp = *mi_eval_scalar(&paras->f_specular1Exponent);
		f_spec2Col = mi_eval_color(&paras->f_specular2Color);
		f_spec2Exp = *mi_eval_scalar(&paras->f_specular2Exponent);
		f_tangentBreakupAm = *mi_eval_scalar(&paras->f_tangentBreakupAmt);
		Ts = UVblend(deriv_u, deriv_v, *mi_eval_scalar(&paras->f_s_angle));
		Ts = shiftTangent(Ts, hairN, f_tangentBreakupAm * mi_noise_1d(bup));
	}

//	coeff_reflect = *mi_eval_scalar(&paras->cReflect);						//Reflect
	if(sp->use_reflection && do_reflect && (coeff_reflect > miEPS) && (alpha > miEPS))
	{
		miColor custom_reflect = *mi_eval_color(&paras->custom_reflect);
//		miScalar fresnel_coeff = 1;
//		if(fresnel_refl_coeff > miEPS)
//			fresnel_coeff = fresnel * fresnel_refl_coeff + (1 - fresnel_refl_coeff);
		if(coeff_reflect > miEPS)
		{
			int reflection_depth = *mi_eval_integer(&paras->reflection_depth);
			miOptions *orig_options = 0;
			miOptions m_options;
			if(reflection_depth > 0 && reflection_depth != state->options->reflection_depth)
			{
				orig_options = (miOptions *)(state->options);		//save pointer to original options
				m_options = *state->options;						//copy original options
				state->options = &m_options;						//link new options to state
				m_options.reflection_depth = reflection_depth;		//override reflection depth
			}
			miColor reflection_color = *mi_eval_color(&paras->glossy_refl);
			reflection_color.r *= coeff_reflect * rr_scale;
			reflection_color.g *= coeff_reflect * rr_scale;
			reflection_color.b *= coeff_reflect * rr_scale;

			if(state->reflection_level <= state->options->reflection_depth && ((reflection_color.r > miEPS) || (reflection_color.g > miEPS) || (reflection_color.b > miEPS)))
				p_reflect(&special.out_reflect, state, paras, &reflection_color, mi_eval_scalar(&paras->shiny_refl),
				*mi_eval_boolean(&paras->anisotropic_reflect), mi_eval_scalar(&paras->shiny_refl_u), mi_eval_scalar(&paras->shiny_refl_v),
				*mi_eval_scalar(&paras->falloff_start_refl), *mi_eval_scalar(&paras->falloff_stop_refl), 
				*mi_eval_boolean(&paras->falloff_to_color_refl), mi_eval_color(&paras->falloff_color_refl),
				*mi_eval_integer(&paras->samples_refl), sp->samples_decrease);
			
			if(orig_options != 0)								//original options was changed
				state->options = orig_options;				//restore pointer to original options

//			float coeff = coeff_reflect * rr_scale;
//			special.out_reflect.r *= rr_scale;
//			special.out_reflect.g *= rr_scale;
//			special.out_reflect.b *= rr_scale;
		}
		special.out_reflect.r += custom_reflect.r * coeff_reflect;
		special.out_reflect.g += custom_reflect.g * coeff_reflect;
		special.out_reflect.b += custom_reflect.b * coeff_reflect;
//		refl_refr += coeff_reflect * (*mi_eval_scalar(&paras->diff_minus_reflect));
	}

//	miScalar coeff_refract;										//Refract
//	coeff_refract = *mi_eval_scalar(&paras->cRefract);
	if(sp->use_refraction && do_refract && (coeff_refract > miEPS) && (alpha > miEPS))
	{
		miColor custom_refract = *mi_eval_color(&paras->custom_refract);
//		miScalar fresnel_coeff = 1;
//		if(fresnel_refr_coeff > miEPS)
//		{
//			fresnel = (1.0f - fresnel);
//			fresnel_coeff = fresnel * fresnel_refr_coeff + (1.0f - fresnel_refr_coeff);
//		}

		int refraction_depth = *mi_eval_integer(&paras->refraction_depth);
		miOptions *orig_options = 0;
		miOptions m_options;
		if(refraction_depth > 0 && refraction_depth != state->options->refraction_depth)
		{
			orig_options = (miOptions *)(state->options);		//save pointer to original options
			m_options = *state->options;						//copy original options
			state->options = &m_options;						//link new options to state
			m_options.refraction_depth = refraction_depth;		//override refraction depth
		}
		miColor refraction_color = *mi_eval_color(&paras->glossy_refr);
		refraction_color.r *= coeff_refract;
		refraction_color.g *= coeff_refract;
		refraction_color.b *= coeff_refract;
		if(state->refraction_level <= state->options->refraction_depth && ((refraction_color.r > miEPS) || (refraction_color.g > miEPS) || (refraction_color.b > miEPS)))
			p_refract(&special.out_refract, state, paras, &refraction_color, mi_eval_scalar(&paras->shiny_refr),
			sp->anisotropic_refract, mi_eval_scalar(&paras->shiny_refr_u), mi_eval_scalar(&paras->shiny_refr_v),
			mi_eval_color(&paras->refr_absorb_in), mi_eval_color(&paras->refr_absorb_out), 
			*mi_eval_scalar(&paras->falloff_start_refr), *mi_eval_scalar(&paras->falloff_stop_refr), 
			*mi_eval_boolean(&paras->falloff_to_color_refr), mi_eval_color(&paras->falloff_color_refr),
			*mi_eval_integer(&paras->samples_refr), enter_surf, sp->samples_decrease);

		if(orig_options != 0)								//original options was changed
			state->options = orig_options;					//restore pointer to original options

		float coeff;
		if (enter_surf)
			coeff = ior_out * ior_out / (ior_in * ior_in) * rr_scale;
		else
			coeff = ior_in * ior_in / (ior_out * ior_out) * rr_scale;

		special.out_refract.r *= coeff;
		special.out_refract.g *= coeff;
		special.out_refract.b *= coeff;
		
		special.out_refract.r += custom_refract.r * coeff_refract;
		special.out_refract.g += custom_refract.g * coeff_refract;
		special.out_refract.b += custom_refract.b * coeff_refract;
//		refl_refr += coeff_refract * (*mi_eval_scalar(&paras->diff_minus_refract));
	}


	if(alpha > miEPS && (sp->use_simple_diffuse || sp->use_fur_diffuse || sp->use_phong_specular || sp->use_blinn_specular || sp->use_velvet_specular || sp->use_wet_specular || sp->use_ward_specular  || sp->use_fur_specular))
	{
		state->normal = orig_normal;
//		void* pri = state->pri;
//		state->pri = NULL;

		/* correct light list, if requested */
		if (mode == 1)
			mi_inclusive_lightlist(&n_light, &light, state);
		else if (mode == 2)
			mi_exclusive_lightlist(&n_light, &light, state);
		else if (mode == 4) 
		{
			n_light = 0;
			light = 0;
		}

		if (mode == 4 || n_light) 
		{
			for (mi::shader::LightIterator iter(state, light, n_light); !iter.at_end(); ++iter) 
			{
				sum_light_shd.r = sum_light_shd.g = sum_light_shd.b = 0;
				sum_light.r = sum_light.g = sum_light.b = 0;
				sum_specular.r = sum_specular.g = sum_specular.b = 0;
				sum_specular_shd.r = sum_specular_shd.g = sum_specular_shd.b = 0;
	//			samples = 0;

	//			state->normal = orig_normal;
	//			state->normal.x = state->normal.y = state->normal.z = 0;

				char *light_category = &light_col.category[0];
		//		while(mi_sample_light(&light_col.l_color_shd, &dir, &dot_nl, state, *light, &samples))
				while (iter->sample()) 
				{
					iter->get_contribution(&light_col.l_color_shd);
					dir = iter->get_direction();
					dot_nl = iter->get_dot_nl();
					if(sp->use_normal)							//if bump, decrease dot_nl close by terminator
					{
						if(state->type == miRAY_FINALGATHER && sp->ind_ignore_bump)
						{}
						else
						{
							state->normal = bump_normal;		//apply bump
		//					geom_dot_nl = mi_vector_dot(&orig_normal, &dir);
							geom_dot_nl = dot_nl;
							dot_nl = mi_vector_dot(&state->normal, &dir);
							if(geom_dot_nl <= 0.2 && dot_nl > 0)
								dot_nl = dot_nl * (geom_dot_nl*5) + geom_dot_nl * (1-(geom_dot_nl*5));
		//					if(geom_dot_nl <= 0.2 && dot_nl > 0)
		//						dot_nl = dot_nl * (geom_dot_nl*5);
							if(dot_nl < 0)
								dot_nl = 0;
						}
					}

//					size_light_color = 0;
					lightTK_pt = (light_color *)mi_shaderstate_get(state, lightsTK_key, NULL);  

					if(lightTK_pt) //Yes, it's LightTK! Good. We compute shading twice, with light only and with light+shadow
					{
						light_col = *lightTK_pt;
						//Diffuse
						if(light_col.emit_diffuse && (refl_refr < 1))
						{
							if(sp->use_simple_diffuse && coeff_dif_simp > miEPS)		//Simple diffuse
							{
								if(light_col.category)
								{
									if(sd_roughness == 1.f)
										dot_nl_rougness = dot_nl;
									else if (dot_nl)
										dot_nl_rougness = pow(dot_nl, sd_roughness);
			
		//							char *categ = &light_col.category[0];
									if((!sp->sd_light_category || !sp->sd_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
									{
		//								mi_warning("p_MegaTK: empty light category specified.");
										sum_light_shd.r += light_col.l_color_shd.r * dot_nl_rougness * coeff_dif_simp;
										sum_light_shd.g += light_col.l_color_shd.g * dot_nl_rougness * coeff_dif_simp;
										sum_light_shd.b += light_col.l_color_shd.b * dot_nl_rougness * coeff_dif_simp;

										sum_light.r += light_col.l_color.r * dot_nl_rougness * coeff_dif_simp;
										sum_light.g += light_col.l_color.g * dot_nl_rougness * coeff_dif_simp;
										sum_light.b += light_col.l_color.b * dot_nl_rougness * coeff_dif_simp;
									}
									else	//some light category specified
									{
										if(strcmp(sp->sd_light_category, light_category) == 0) //same light categories
										{
		//									mi_warning("p_MegaTK: same light categories.");
											sum_light_shd.r += light_col.l_color_shd.r * dot_nl_rougness * coeff_dif_simp;
											sum_light_shd.g += light_col.l_color_shd.g * dot_nl_rougness * coeff_dif_simp;
											sum_light_shd.b += light_col.l_color_shd.b * dot_nl_rougness * coeff_dif_simp;

											sum_light.r += light_col.l_color.r * dot_nl_rougness * coeff_dif_simp;
											sum_light.g += light_col.l_color.g * dot_nl_rougness * coeff_dif_simp;
											sum_light.b += light_col.l_color.b * dot_nl_rougness * coeff_dif_simp;
										}
									}
								}
							}
							if(sp->use_fur_diffuse && (coeff_dif_fur > 0))		//Fur diffuse
							{
								if((!sp->fd_light_category || !sp->fd_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
								{
									miColor fur_dif_shd = p_fur_diffuse(f_skinCol, f_hairDens, f_backwardScater,
										f_forwardScatter, f_hairOverHairShad, f_shadowAtten, f_startShadowCo,
										f_endShadowCo, hair_area, &hairN, &V, &T, &dir, &light_col.l_color_shd, dot_nl);
									miColor fur_dif = p_fur_diffuse(f_skinCol, f_hairDens, f_backwardScater,
										f_forwardScatter, f_hairOverHairShad, f_shadowAtten, f_startShadowCo,
										f_endShadowCo, hair_area, &hairN, &V, &T, &dir, &light_col.l_color_shd, dot_nl);
									fur_dif_shd.r *= coeff_dif_fur;
									fur_dif_shd.g *= coeff_dif_fur;
									fur_dif_shd.b *= coeff_dif_fur;

									fur_dif.r *= coeff_dif_fur;
									fur_dif.g *= coeff_dif_fur;
									fur_dif.b *= coeff_dif_fur;

									sum_light_shd.r += fur_dif_shd.r;
									sum_light_shd.g += fur_dif_shd.g;
									sum_light_shd.b += fur_dif_shd.b;

									sum_light.r += fur_dif.r;
									sum_light.g += fur_dif.g;
									sum_light.b += fur_dif.b;
								}
								else	//some light category specified
								{
									if(strcmp(sp->fd_light_category, light_category) == 0) //same light categories
									{
		//								mi_warning("p_MegaTK: same light categories.");
										miColor fur_dif_shd = p_fur_diffuse(f_skinCol, f_hairDens, f_backwardScater,
											f_forwardScatter, f_hairOverHairShad, f_shadowAtten, f_startShadowCo,
											f_endShadowCo, hair_area, &hairN, &V, &T, &dir, &light_col.l_color_shd, dot_nl);
										miColor fur_dif = p_fur_diffuse(f_skinCol, f_hairDens, f_backwardScater,
											f_forwardScatter, f_hairOverHairShad, f_shadowAtten, f_startShadowCo,
											f_endShadowCo, hair_area, &hairN, &V, &T, &dir, &light_col.l_color_shd, dot_nl);
										fur_dif_shd.r *= coeff_dif_fur;
										fur_dif_shd.g *= coeff_dif_fur;
										fur_dif_shd.b *= coeff_dif_fur;

										fur_dif.r *= coeff_dif_fur;
										fur_dif.g *= coeff_dif_fur;
										fur_dif.b *= coeff_dif_fur;

										sum_light_shd.r += fur_dif_shd.r;
										sum_light_shd.g += fur_dif_shd.g;
										sum_light_shd.b += fur_dif_shd.b;

										sum_light.r += fur_dif.r;
										sum_light.g += fur_dif.g;
										sum_light.b += fur_dif.b;
									}
								}
							}
						}

						//Specular
						if(light_col.emit_specular)
						{
							if(sp->use_phong_specular && (*p_coeffS > 0))	//Phong specular
							{
								if((!sp->p_light_category || !sp->p_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
								{
									s = mi_phong_specular(expo, state, &dir);
									if (s > 0.0)
									{
										sum_specular_shd.r += s * p_spec->r * *p_coeffS * light_col.l_color_shd.r;
										sum_specular_shd.g += s * p_spec->g * *p_coeffS * light_col.l_color_shd.g;
										sum_specular_shd.b += s * p_spec->b * *p_coeffS * light_col.l_color_shd.b;

										sum_specular.r += s * p_spec->r * *p_coeffS * light_col.l_color.r;
										sum_specular.g += s * p_spec->g * *p_coeffS * light_col.l_color.g;
										sum_specular.b += s * p_spec->b * *p_coeffS * light_col.l_color.b;
									}
								}
								else	//some light category specified
								{
									if(strcmp(sp->p_light_category, light_category) == 0) //same light categories
									{
		//								mi_warning("p_MegaTK: same light categories.");
										s = mi_phong_specular(expo, state, &dir);
										if (s > 0.0)
										{
											sum_specular_shd.r += s * p_spec->r * *p_coeffS * light_col.l_color_shd.r;
											sum_specular_shd.g += s * p_spec->g * *p_coeffS * light_col.l_color_shd.g;
											sum_specular_shd.b += s * p_spec->b * *p_coeffS * light_col.l_color_shd.b;

											sum_specular.r += s * p_spec->r * *p_coeffS * light_col.l_color.r;
											sum_specular.g += s * p_spec->g * *p_coeffS * light_col.l_color.g;
											sum_specular.b += s * p_spec->b * *p_coeffS * light_col.l_color.b;
										}
									}
								}
							}
							if(sp->use_blinn_specular && (*b_coeffS > 0))	//Blinn specular
							{
								if((!sp->b_light_category || !sp->b_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
								{
	//								mi_warning("p_MegaTK: empty light categories.");
									Ln = dir;
									mi_vector_normalize(&Ln);
									mi_vector_add(&H, &Ln, &V);
									mi_vector_normalize(&H);
									NH = mi_vector_dot(&state->normal, &H);
									NHSQ = NH * NH;
									NH2 = NH * 2;
									Dd = (E+1) / (NHSQ + E);
									Dd *= Dd;
									VH = mi_vector_dot(&V, &H);
									LN = mi_vector_dot(&Ln, &state->normal);
									if(VN < LN)
									{
										if((VN * NH2) < VH )
											Gg = NH2 / VH;
										else
											Gg = 1 / VN;
									}
									else
									{
										if((LN * NH2) < VH )
											Gg = (LN * NH2) / (VH * VN);
										else
											Gg = 1 / VN;
									}
									// poor man's Fresnel
									tmp = pow((1 - VH), 3);
									Ff = tmp + (1 - tmp) * b_rolloff;
									s = Dd * Gg * Ff;
									if (s > 0.0)
									{
										sum_specular_shd.r += s * b_spec->r * *b_coeffS * light_col.l_color_shd.r;
										sum_specular_shd.g += s * b_spec->g * *b_coeffS * light_col.l_color_shd.g;
										sum_specular_shd.b += s * b_spec->b * *b_coeffS * light_col.l_color_shd.b;

										sum_specular.r += s * b_spec->r * *b_coeffS * light_col.l_color.r;
										sum_specular.g += s * b_spec->g * *b_coeffS * light_col.l_color.g;
										sum_specular.b += s * b_spec->b * *b_coeffS * light_col.l_color.b;
									}
								}
								else	//some light category specified
								{
									if(strcmp(sp->b_light_category, light_category) == 0) //same light categories
									{
				//						mi_warning("p_MegaTK: same light categories.");
										Ln = dir;
										mi_vector_normalize(&Ln);
										mi_vector_add(&H, &Ln, &V);
										mi_vector_normalize(&H);
										NH = mi_vector_dot(&state->normal, &H);
										NHSQ = NH * NH;
										NH2 = NH * 2;
										Dd = (E+1) / (NHSQ + E);
										Dd *= Dd;
										VH = mi_vector_dot(&V, &H);
										LN = mi_vector_dot(&Ln, &state->normal);
										if(VN < LN)
										{
											if((VN * NH2) < VH )
												Gg = NH2 / VH;
											else
												Gg = 1 / VN;
										}
										else
										{
											if((LN * NH2) < VH )
												Gg = (LN * NH2) / (VH * VN);
											else
												Gg = 1 / VN;
										}
										// poor man's Fresnel
										tmp = pow((1 - VH), 3);
										Ff = tmp + (1 - tmp) * b_rolloff;
										s = Dd * Gg * Ff;
										if (s > 0.0)
										{
											sum_specular_shd.r += s * b_spec->r * *b_coeffS * light_col.l_color_shd.r;
											sum_specular_shd.g += s * b_spec->g * *b_coeffS * light_col.l_color_shd.g;
											sum_specular_shd.b += s * b_spec->b * *b_coeffS * light_col.l_color_shd.b;

											sum_specular.r += s * b_spec->r * *b_coeffS * light_col.l_color.r;
											sum_specular.g += s * b_spec->g * *b_coeffS * light_col.l_color.g;
											sum_specular.b += s * b_spec->b * *b_coeffS * light_col.l_color.b;
										}
									}
								}
							}
							if(sp->use_velvet_specular && (*v_coeffS > 0))	//Velvet specular
							{
								if((!sp->v_light_category || !sp->v_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
								{
									miScalar temp_vector_dot = mi_vector_dot(&dir, &V);
									cosine = (temp_vector_dot > 0.0f ? temp_vector_dot : 0.0f);
									sum_specular_shd.r += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.r * v_sheen_color->r * *v_backscat);
									sum_specular_shd.g += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.g * v_sheen_color->g * *v_backscat);
									sum_specular_shd.b += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.b * v_sheen_color->b * *v_backscat);

									sum_specular_shd.r += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.r * v_sheen_color->r * dot_nl);
									sum_specular_shd.g += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.g * v_sheen_color->g * dot_nl);
									sum_specular_shd.b += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.b * v_sheen_color->b * dot_nl);

									sum_specular.r += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color.r * v_sheen_color->r * *v_backscat);
									sum_specular.g += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color.g * v_sheen_color->g * *v_backscat);
									sum_specular.b += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color.b * v_sheen_color->b * *v_backscat);

									sum_specular.r += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color.r * v_sheen_color->r * dot_nl);
									sum_specular.g += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color.g * v_sheen_color->g * dot_nl);
									sum_specular.b += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color.b * v_sheen_color->b * dot_nl);
								}
								else	//some light category specified
								{
									if(strcmp(sp->v_light_category, light_category) == 0) //same light categories
									{
		//								mi_warning("p_MegaTK: same light categories.");
										miScalar temp_vector_dot = mi_vector_dot(&dir, &V);
										cosine = (temp_vector_dot > 0.0f ? temp_vector_dot : 0.0f);
										sum_specular_shd.r += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.r * v_sheen_color->r * *v_backscat);
										sum_specular_shd.g += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.g * v_sheen_color->g * *v_backscat);
										sum_specular_shd.b += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.b * v_sheen_color->b * *v_backscat);

										sum_specular_shd.r += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.r * v_sheen_color->r * dot_nl);
										sum_specular_shd.g += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.g * v_sheen_color->g * dot_nl);
										sum_specular_shd.b += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.b * v_sheen_color->b * dot_nl);

										sum_specular.r += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color.r * v_sheen_color->r * *v_backscat);
										sum_specular.g += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color.g * v_sheen_color->g * *v_backscat);
										sum_specular.b += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color.b * v_sheen_color->b * *v_backscat);

										sum_specular.r += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color.r * v_sheen_color->r * dot_nl);
										sum_specular.g += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color.g * v_sheen_color->g * dot_nl);
										sum_specular.b += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color.b * v_sheen_color->b * dot_nl);
									}
								}
							}
							if(sp->use_wet_specular && (we_cS > 0))	//Wet specular
							{
								if((!sp->we_light_category || !sp->we_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
								{
									Ln = dir;
									mi_vector_normalize(&Ln);
									mi_vector_add(&H, &Ln, &V);
									mi_vector_normalize(&H);
									NH = mi_vector_dot(&state->normal, &H);

									we_falloff = dot_nl * 1.2f;
									if(we_falloff < 0)
										NH = 0;					//beside the light terminator
									else if (we_falloff < 1)
									{
										we_falloff = pow(we_falloff, 0.05f);
										NH -= 1-we_falloff;		//specular offset close by light terminator
									}

									if(NH > 0)
									{							
										if(NH > we_roughness)
										{
											sum_specular_shd.r += we_specular->r * we_cS * light_col.l_color_shd.r;
											sum_specular_shd.g += we_specular->g * we_cS * light_col.l_color_shd.g;
											sum_specular_shd.b += we_specular->b * we_cS * light_col.l_color_shd.b;

											sum_specular.r += we_specular->r * we_cS * light_col.l_color.r;
											sum_specular.g += we_specular->g * we_cS * light_col.l_color.g;
											sum_specular.b += we_specular->b * we_cS * light_col.l_color.b;
										}
										else if (NH > (we_roughness - we_sharpness))
										{
											we_value = (NH -(we_roughness - we_sharpness)) / we_sharpness * we_cS;
											if(we_cS > 0)
											{
												sum_specular_shd.r += we_specular->r * we_value * light_col.l_color_shd.r;
												sum_specular_shd.g += we_specular->g * we_value * light_col.l_color_shd.g;
												sum_specular_shd.b += we_specular->b * we_value * light_col.l_color_shd.b;

												sum_specular.r += we_specular->r * we_value * light_col.l_color.r;
												sum_specular.g += we_specular->g * we_value * light_col.l_color.g;
												sum_specular.b += we_specular->b * we_value * light_col.l_color.b;
											}
										}
									}
								}
								else if(strcmp(sp->we_light_category, light_category) == 0) //same light categories
								{
									Ln = dir;
									mi_vector_normalize(&Ln);
									mi_vector_add(&H, &Ln, &V);
									mi_vector_normalize(&H);
									NH = mi_vector_dot(&state->normal, &H);

									we_falloff = dot_nl * 1.2f;
									if(we_falloff < 0)
										NH = 0;					//beside the light terminator
									else if (we_falloff < 1)
									{
										we_falloff = pow(we_falloff, 0.05f);
										NH -= 1-we_falloff;		//specular offset close by light terminator
									}

									if(NH > 0)
									{							
										if(NH > we_roughness)
										{
											sum_specular_shd.r += we_specular->r * we_cS * light_col.l_color_shd.r;
											sum_specular_shd.g += we_specular->g * we_cS * light_col.l_color_shd.g;
											sum_specular_shd.b += we_specular->b * we_cS * light_col.l_color_shd.b;

											sum_specular.r += we_specular->r * we_cS * light_col.l_color.r;
											sum_specular.g += we_specular->g * we_cS * light_col.l_color.g;
											sum_specular.b += we_specular->b * we_cS * light_col.l_color.b;
										}
										else if (NH > (we_roughness - we_sharpness))
										{
											we_value = (NH -(we_roughness - we_sharpness)) / we_sharpness * we_cS;
											if(we_cS > 0)
											{
												sum_specular_shd.r += we_specular->r * we_value * light_col.l_color_shd.r;
												sum_specular_shd.g += we_specular->g * we_value * light_col.l_color_shd.g;
												sum_specular_shd.b += we_specular->b * we_value * light_col.l_color_shd.b;

												sum_specular.r += we_specular->r * we_value * light_col.l_color.r;
												sum_specular.g += we_specular->g * we_value * light_col.l_color.g;
												sum_specular.b += we_specular->b * we_value * light_col.l_color.b;
											}
										}
									}
								}
							}
							if(sp->use_ward_specular && (w_coeffS > 0))			//Ward specular
							{
								if((!sp->w_light_category || !sp->w_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
								{
									if (w_shin_u == w_shin_v)		// isotropic refl.
										w_spec = w_coeffS * dot_nl * mi_ward_glossy(&state->dir, &dir, &state->normal, w_shin_u);
									else							// anisotropic refl.
									{
										if(sp->w_use_deriv)
										{
											miVector	w_u;
											miVector	w_v;
											float		w_d;
											w_u = state->derivs[0];
											w_d = mi_vector_dot(&w_u, &state->normal);
											w_u.x -= w_d * state->normal.x;
											w_u.y -= w_d * state->normal.y;
											w_u.z -= w_d * state->normal.z;
											mi_vector_normalize(&w_u);
											mi_vector_prod(&w_v, &state->normal, &w_u);
											w_spec = w_coeffS * dot_nl * mi_ward_anisglossy(&state->dir, &dir, &state->normal, &w_u, &w_v, w_shin_u, w_shin_v);
										}
										else
										{
											miVector *w_u = mi_eval_vector(&paras->w_u);
											miVector *w_v = mi_eval_vector(&paras->w_v);
											w_spec = w_coeffS * dot_nl * mi_ward_anisglossy(&state->dir, &dir, &state->normal, w_u, w_v, w_shin_u, w_shin_v);
										}
									}
									if(w_spec > 0.0)
									{
										sum_specular_shd.r += w_spec * w_glos->r * light_col.l_color_shd.r;
										sum_specular_shd.g += w_spec * w_glos->g * light_col.l_color_shd.g;
										sum_specular_shd.b += w_spec * w_glos->b * light_col.l_color_shd.b;

										sum_specular.r += w_spec * w_glos->r * light_col.l_color.r;
										sum_specular.g += w_spec * w_glos->g * light_col.l_color.g;
										sum_specular.b += w_spec * w_glos->b * light_col.l_color.b;
									}
								}
								else	//some light category specified
								{
									if(strcmp(sp->w_light_category, light_category) == 0) //same light categories
									{
		//								mi_warning("p_MegaTK: same light categories.");
										if (w_shin_u == w_shin_v)		// isotropic refl.
											w_spec = w_coeffS * dot_nl * mi_ward_glossy(&state->dir, &dir, &state->normal, w_shin_u);
										else							// anisotropic refl.
										{
											if(sp->w_use_deriv)
											{
												miVector	w_u;
												miVector	w_v;
												float		w_d;
												w_u = state->derivs[0];
												w_d = mi_vector_dot(&w_u, &state->normal);
												w_u.x -= w_d * state->normal.x;
												w_u.y -= w_d * state->normal.y;
												w_u.z -= w_d * state->normal.z;
												mi_vector_normalize(&w_u);
												mi_vector_prod(&w_v, &state->normal, &w_u);
												w_spec = w_coeffS * dot_nl * mi_ward_anisglossy(&state->dir, &dir, &state->normal, &w_u, &w_v, w_shin_u, w_shin_v);
											}
											else
											{
												miVector *w_u = mi_eval_vector(&paras->w_u);
												miVector *w_v = mi_eval_vector(&paras->w_v);
												w_spec = w_coeffS * dot_nl * mi_ward_anisglossy(&state->dir, &dir, &state->normal, w_u, w_v, w_shin_u, w_shin_v);
											}
										}
										if(w_spec > 0.0)
										{
											sum_specular_shd.r += w_spec * w_glos->r * light_col.l_color_shd.r;
											sum_specular_shd.g += w_spec * w_glos->g * light_col.l_color_shd.g;
											sum_specular_shd.b += w_spec * w_glos->b * light_col.l_color_shd.b;

											sum_specular.r += w_spec * w_glos->r * light_col.l_color.r;
											sum_specular.g += w_spec * w_glos->g * light_col.l_color.g;
											sum_specular.b += w_spec * w_glos->b * light_col.l_color.b;
										}
									}
								}
							}

							if(sp->use_fur_specular && (f_coeffS > 0))			//Fur specular
							{
								if((!sp->v_light_category || !sp->v_light_category[0]) && (!light_category || !light_category[0]))  //empty light category specified
								{
									miColor fur_result_shd = p_fur_specular(f_spec1Col, f_spec1Exp, f_spec2Col, f_spec2Exp,
										&hairN, &V, &Ts, &dir, &light_col.l_color_shd);
									miColor fur_result = p_fur_specular(f_spec1Col, f_spec1Exp, f_spec2Col, f_spec2Exp,
										&hairN, &V, &Ts, &dir, &light_col.l_color);
									fur_result_shd.r *= f_coeffS * dot_nl;
									fur_result_shd.g *= f_coeffS * dot_nl;
									fur_result_shd.b *= f_coeffS * dot_nl;

									fur_result.r *= f_coeffS * dot_nl;
									fur_result.g *= f_coeffS * dot_nl;
									fur_result.b *= f_coeffS * dot_nl;

									sum_specular_shd.r += fur_result_shd.r;
									sum_specular_shd.g += fur_result_shd.g;
									sum_specular_shd.b += fur_result_shd.b;

									sum_specular.r += fur_result.r;
									sum_specular.g += fur_result.g;
									sum_specular.b += fur_result.b;
								}
								else	//some light category specified
								{
									if(strcmp(sp->v_light_category, light_category) == 0) //same light categories
									{
		//								mi_warning("p_MegaTK: same light categories.");
										miColor fur_result_shd = p_fur_specular(f_spec1Col, f_spec1Exp, f_spec2Col, f_spec2Exp,
											&hairN, &V, &Ts, &dir, &light_col.l_color_shd);
										miColor fur_result = p_fur_specular(f_spec1Col, f_spec1Exp, f_spec2Col, f_spec2Exp,
											&hairN, &V, &Ts, &dir, &light_col.l_color);
										fur_result_shd.r *= f_coeffS * dot_nl;
										fur_result_shd.g *= f_coeffS * dot_nl;
										fur_result_shd.b *= f_coeffS * dot_nl;

										fur_result.r *= f_coeffS * dot_nl;
										fur_result.g *= f_coeffS * dot_nl;
										fur_result.b *= f_coeffS * dot_nl;

										sum_specular_shd.r += fur_result_shd.r;
										sum_specular_shd.g += fur_result_shd.g;
										sum_specular_shd.b += fur_result_shd.b;

										sum_specular.r += fur_result.r;
										sum_specular.g += fur_result.g;
										sum_specular.b += fur_result.b;
									}
								}
							}
						}
					}
					else				//It's not LightTK! Bad, but should work. It's work like any other common shader.
					{
						//Diffuse
						if(refl_refr < 1)
						{
							if(sp->use_simple_diffuse && (coeff_dif_simp > miEPS))		//Simple diffuse
							{
									if(sd_roughness == 1.f)
										dot_nl_rougness = dot_nl;
									else if (dot_nl)
										dot_nl_rougness = pow(dot_nl, sd_roughness);

									sum_light_shd.r += light_col.l_color_shd.r * dot_nl_rougness * coeff_dif_simp;
									sum_light_shd.g += light_col.l_color_shd.g * dot_nl_rougness * coeff_dif_simp;
									sum_light_shd.b += light_col.l_color_shd.b * dot_nl_rougness * coeff_dif_simp;
							}

							if(sp->use_fur_diffuse && (coeff_dif_fur > 0))		//Fur diffuse
							{

								miColor fur_dif_shd = p_fur_diffuse(f_skinCol, f_hairDens, f_backwardScater,
									f_forwardScatter, f_hairOverHairShad, f_shadowAtten, f_startShadowCo,
									f_endShadowCo, hair_area, &hairN, &V, &T, &dir, &light_col.l_color_shd, dot_nl);
								fur_dif_shd.r *= coeff_dif_fur;
								fur_dif_shd.g *= coeff_dif_fur;
								fur_dif_shd.b *= coeff_dif_fur;

								sum_light_shd.r += fur_dif_shd.r;
								sum_light_shd.g += fur_dif_shd.g;
								sum_light_shd.b += fur_dif_shd.b;
							}
						}

						//Specular

						if(sp->use_phong_specular && (*p_coeffS > 0))	//Phong specular
						{
							s = mi_phong_specular(expo, state, &dir);
							if (s > 0.0)
							{
								sum_specular_shd.r += s * p_spec->r * *p_coeffS * light_col.l_color_shd.r;
								sum_specular_shd.g += s * p_spec->g * *p_coeffS * light_col.l_color_shd.g;
								sum_specular_shd.b += s * p_spec->b * *p_coeffS * light_col.l_color_shd.b;
							}
						}

						if(sp->use_blinn_specular && (*b_coeffS > 0))	//Blinn specular
						{
							Ln = dir;
							mi_vector_normalize(&Ln);
							mi_vector_add(&H, &Ln, &V);
							mi_vector_normalize(&H);
							NH = mi_vector_dot(&state->normal, &H);
							NHSQ = NH * NH;
							NH2 = NH * 2;
							Dd = (E+1) / (NHSQ + E);
							Dd *= Dd;
							VH = mi_vector_dot(&V, &H);
							LN = mi_vector_dot(&Ln, &state->normal);
							if(VN < LN)
							{
								if((VN * NH2) < VH )
									Gg = NH2 / VH;
								else
									Gg = 1 / VN;
							}
							else
							{
								if((LN * NH2) < VH )
									Gg = (LN * NH2) / (VH * VN);
								else
									Gg = 1 / VN;
							}
							// poor man's Fresnel
							tmp = pow((1 - VH), 3);
							Ff = tmp + (1 - tmp) * b_rolloff;
							s = Dd * Gg * Ff;
							if (s > 0.0)
							{
								sum_specular_shd.r += s * b_spec->r * *b_coeffS * light_col.l_color_shd.r;
								sum_specular_shd.g += s * b_spec->g * *b_coeffS * light_col.l_color_shd.g;
								sum_specular_shd.b += s * b_spec->b * *b_coeffS * light_col.l_color_shd.b;
							}
						}

						if(sp->use_velvet_specular && (*v_coeffS > 0))	//Velvet specular
						{
							miScalar temp_vector_dot = mi_vector_dot(&dir, &V);
							cosine = (temp_vector_dot > 0.0f ? temp_vector_dot : 0.0f);

							sum_specular_shd.r += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.r * v_sheen_color->r * *v_backscat);
							sum_specular_shd.g += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.g * v_sheen_color->g * *v_backscat);
							sum_specular_shd.b += (pow( cosine, iRoughness ) * *v_coeffS * light_col.l_color_shd.b * v_sheen_color->b * *v_backscat);

							sum_specular_shd.r += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.r * v_sheen_color->r * dot_nl);
							sum_specular_shd.g += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.g * v_sheen_color->g * dot_nl);
							sum_specular_shd.b += (pow( sine, *v_edgin ) * *v_coeffS * light_col.l_color_shd.b * v_sheen_color->b * dot_nl);
						}

						if(sp->use_wet_specular && (we_cS > 0))	//Wet specular
						{
							Ln = dir;
							mi_vector_normalize(&Ln);
							mi_vector_add(&H, &Ln, &V);
							mi_vector_normalize(&H);
							NH = mi_vector_dot(&state->normal, &H);

							we_falloff = dot_nl * 1.2f;
							if(we_falloff < 0)
								NH = 0;					//beside the light terminator
							else if (we_falloff < 1)
							{
								we_falloff = pow(we_falloff, 0.05f);
								NH -= 1-we_falloff;		//specular offset close by light terminator
							}

							if(NH > 0)
							{							
								if(NH > we_roughness)
								{
									sum_specular_shd.r += we_specular->r * we_cS * light_col.l_color_shd.r;
									sum_specular_shd.g += we_specular->g * we_cS * light_col.l_color_shd.g;
									sum_specular_shd.b += we_specular->b * we_cS * light_col.l_color_shd.b;
								}
								else if (NH > (we_roughness - we_sharpness))
								{
									we_value = (NH -(we_roughness - we_sharpness)) / we_sharpness * we_cS;
									if(we_cS > 0)
									{
										sum_specular_shd.r += we_specular->r * we_value * light_col.l_color_shd.r;
										sum_specular_shd.g += we_specular->g * we_value * light_col.l_color_shd.g;
										sum_specular_shd.b += we_specular->b * we_value * light_col.l_color_shd.b;
									}
								}
							}
						}

						if(sp->use_ward_specular && (w_coeffS > 0))			//Ward specular
						{
							if (w_shin_u == w_shin_v)		// isotropic refl.
								w_spec = w_coeffS * dot_nl * mi_ward_glossy(&state->dir, &dir, &state->normal, w_shin_u);
							else							// anisotropic refl.
							{
								if(sp->w_use_deriv)
								{
									miVector	w_u;
									miVector	w_v;
									float		w_d;
									w_u = state->derivs[0];
									w_d = mi_vector_dot(&w_u, &state->normal);
									w_u.x -= w_d * state->normal.x;
									w_u.y -= w_d * state->normal.y;
									w_u.z -= w_d * state->normal.z;
									mi_vector_normalize(&w_u);
									mi_vector_prod(&w_v, &state->normal, &w_u);
									w_spec = w_coeffS * dot_nl * mi_ward_anisglossy(&state->dir, &dir, &state->normal, &w_u, &w_v, w_shin_u, w_shin_v);
								}
								else
								{
									miVector *w_u = mi_eval_vector(&paras->w_u);
									miVector *w_v = mi_eval_vector(&paras->w_v);
									w_spec = w_coeffS * dot_nl * mi_ward_anisglossy(&state->dir, &dir, &state->normal, w_u, w_v, w_shin_u, w_shin_v);
								}
							}
							if(w_spec > 0.0)
							{
								sum_specular_shd.r += w_spec * w_glos->r * light_col.l_color_shd.r;
								sum_specular_shd.g += w_spec * w_glos->g * light_col.l_color_shd.g;
								sum_specular_shd.b += w_spec * w_glos->b * light_col.l_color_shd.b;
							}
						}

						if(sp->use_fur_specular && (f_coeffS > 0))			//Fur specular
						{
							miColor fur_result_shd = p_fur_specular(f_spec1Col, f_spec1Exp, f_spec2Col, f_spec2Exp,
								&hairN, &V, &Ts, &dir, &light_col.l_color_shd);
							miColor fur_result = p_fur_specular(f_spec1Col, f_spec1Exp, f_spec2Col, f_spec2Exp,
								&hairN, &V, &Ts, &dir, &light_col.l_color);
							fur_result_shd.r *= f_coeffS * dot_nl;
							fur_result_shd.g *= f_coeffS * dot_nl;
							fur_result_shd.b *= f_coeffS * dot_nl;

							sum_specular_shd.r += fur_result_shd.r;
							sum_specular_shd.g += fur_result_shd.g;
							sum_specular_shd.b += fur_result_shd.b;
						}
					}

					state->normal = orig_normal;
				}
//				mi_shaderstate_set(state, "lig_tk", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);
				if(lightTK_pt)
					mi_shaderstate_set(state, lightsTK_key, NULL, 0, miSS_LIFETIME_EYERAY);
				
				samples = iter->get_number_of_samples();
				if (samples)
				{
					float inv_samples = 1.f/samples;
					sum_light_shd.r *= inv_samples;
					sum_light_shd.g *= inv_samples;
					sum_light_shd.b *= inv_samples;

					sum_light.r *= inv_samples;
					sum_light.g *= inv_samples;
					sum_light.b *= inv_samples;

					sum_specular.r *= inv_samples;
					sum_specular.g *= inv_samples;
					sum_specular.b *= inv_samples;

					sum_specular_shd.r *= inv_samples;
					sum_specular_shd.g *= inv_samples;
					sum_specular_shd.b *= inv_samples;
				}
				special.out_diffuse_shadow.r += sum_light_shd.r;
				special.out_diffuse_shadow.g += sum_light_shd.g;
				special.out_diffuse_shadow.b += sum_light_shd.b;
				special.out_diffuse.r += sum_light.r;
				special.out_diffuse.g += sum_light.g;
				special.out_diffuse.b += sum_light.b;
				special.out_specular_shadow.r += sum_specular_shd.r;
				special.out_specular_shadow.g += sum_specular_shd.g;
				special.out_specular_shadow.b += sum_specular_shd.b;
				special.out_specular.r += sum_specular.r;
				special.out_specular.g += sum_specular.g;
				special.out_specular.b += sum_specular.b;
			} //for LightIterator
//		state->pri = pri;
		}
		if(sp->use_normal)				//apply bump
		{
			if(state->type == miRAY_FINALGATHER && sp->ind_ignore_bump)
			{}
			else
				state->normal = bump_normal;
		}
	}



//	if((0 < refl_refr) && (refl_refr < 1))							//Minus diffuse
//	{
//		special.out_diffuse_shadow.r *= (1 - refl_refr);
//		special.out_diffuse_shadow.g *= (1 - refl_refr);
//		special.out_diffuse_shadow.b *= (1 - refl_refr);
//		special.out_diffuse.r *= (1 - refl_refr);
//		special.out_diffuse.g *= (1 - refl_refr);
//		special.out_diffuse.b *= (1 - refl_refr);
//	}


	if(sp->use_occlusion && (alpha > miEPS))									//Occlusion
	{
		special.out_occlusion = *mi_eval_color(&paras->custom_occlusion);
		if((state->type == miRAY_REFLECT && sp->o_inv_refl) || (state->type == miRAY_REFRACT && sp->o_inv_refr))
		{}
		else 
		{
			if(sp->o_samples > 0)
			{
				miColor o_result;
				miVector o_bent_result;
				o_bent_result.x = o_bent_result.y = o_bent_result.z = 0;
				p_occlusion(&o_result, &o_bent_result, state, sp, mi_eval_color(&paras->o_base_color), mi_eval_color(&paras->o_dirt_color), *mi_eval_vector(&paras->o_basis), *mi_eval_scalar(&paras->o_basis_blend));
				special.out_occlusion.r *= o_result.r;
				special.out_occlusion.g *= o_result.g;
				special.out_occlusion.b *= o_result.b;
				if(sp->o_bent_normal)
				{
					special.out_bent_normal.r = o_bent_result.x;
					special.out_bent_normal.g = o_bent_result.y;
					special.out_bent_normal.b = o_bent_result.z;
				}
			}
		}
	}


	miScalar coeffAmb = *mi_eval_scalar(&paras->cAmb);							//Ambient
	if(sp->use_ambient && (coeffAmb > miEPS) && (alpha > miEPS))
	{
		miColor *amb_color = mi_eval_color(&paras->ambient_color);
		special.out_ambient.r = special.out_color.r * amb_color->r * coeffAmb;
		special.out_ambient.g = special.out_color.g * amb_color->g * coeffAmb;
		special.out_ambient.b = special.out_color.b * amb_color->b * coeffAmb;

		miScalar * effect_ambient;												//Effect occlusion to ambient
		effect_ambient = mi_eval_scalar(&paras->effect_ambient);
		if(sp->use_occlusion && (*effect_ambient > 0))
		{
			if(*effect_ambient == 1)
			{
				special.out_amb_occlusion.r = special.out_ambient.r * special.out_occlusion.r;
				special.out_amb_occlusion.g = special.out_ambient.g * special.out_occlusion.g;
				special.out_amb_occlusion.b = special.out_ambient.b * special.out_occlusion.b;
			}
			else
			{
				special.out_amb_occlusion.r = special.out_ambient.r * (special.out_occlusion.r * *effect_ambient + (1 - *effect_ambient));
				special.out_amb_occlusion.g = special.out_ambient.g * (special.out_occlusion.g * *effect_ambient + (1 - *effect_ambient));
				special.out_amb_occlusion.b = special.out_ambient.b * (special.out_occlusion.b * *effect_ambient + (1 - *effect_ambient));
			}
		}
		else
		{
			special.out_amb_occlusion.r = special.out_ambient.r;
			special.out_amb_occlusion.g = special.out_ambient.g;
			special.out_amb_occlusion.b = special.out_ambient.b;
		}
	}


	if(sp->use_indirect_illum && (alpha > miEPS))									//Indirect_Illum
	{
		if(sp->use_normal && sp->ind_ignore_bump)
			state->normal = orig_normal; 

		miColor * coeffInd;
		coeffInd = mi_eval_color(&paras->cInd);

		if(sp->path_indirect == miFALSE)
			mi_compute_irradiance(&color, state);
		else
			p_path_diffuse(&color, state, sp->path_indirect_depth, sp->path_indirect_samples, sp->samples_decrease);
		miScalar satur = *mi_eval_scalar(&paras->saturation_indirect);
//		miScalar gray = color.r*0.299f + color.g*0.587f + color.b*0.114f;
		miScalar gray = (color.r + color.g + color.b) * 0.3333f;

		if(satur < 1)
		{
			if(satur == 0)
				color.r = color.g = color.b = gray;
			else
			{
				color.r = gray*(1-satur) + color.r*satur;
				color.g = gray*(1-satur) + color.g*satur;
				color.b = gray*(1-satur) + color.b*satur;
			}
		}

		color.r *= 1-refl_refr;
		color.g *= 1-refl_refr;
		color.b *= 1-refl_refr;


		float effect_indir = *mi_eval_scalar(&paras->effect_indirect);			//Effect occlusion to indirect
		if(sp->use_occlusion && effect_indir > 0)
		{
			if(effect_indir == 1)
			{
				special.out_indirect_illum.r = color.r * special.out_color.r * coeffInd->r * special.out_occlusion.r;
				special.out_indirect_illum.g = color.g * special.out_color.g * coeffInd->g * special.out_occlusion.g;
				special.out_indirect_illum.b = color.b * special.out_color.b * coeffInd->b * special.out_occlusion.b;
			}
			else
			{
				special.out_indirect_illum.r = special.out_color.r * coeffInd->r * (color.r * special.out_occlusion.r * effect_indir + (1 - effect_indir) * color.r);
				special.out_indirect_illum.g = special.out_color.g * coeffInd->g * (color.g * special.out_occlusion.g * effect_indir + (1 - effect_indir) * color.g);
				special.out_indirect_illum.b = special.out_color.b * coeffInd->b * (color.b * special.out_occlusion.b * effect_indir + (1 - effect_indir) * color.b);
			}
		}
		else
		{
				special.out_indirect_illum.r = color.r * special.out_color.r * coeffInd->r;
				special.out_indirect_illum.g = color.g * special.out_color.g * coeffInd->g;
				special.out_indirect_illum.b = color.b * special.out_color.b * coeffInd->b;
		}
		if(sp->use_normal && sp->ind_ignore_bump)
			state->normal = bump_normal; 
	}

											//Effect occlusion to specular
	float effect_spec = *mi_eval_scalar(&paras->effect_specular);
	if(sp->use_occlusion && effect_spec > miEPS)
	{
		if(effect_spec == 1)
		{
			special.out_specular_shadow.r *= special.out_occlusion.r;
			special.out_specular_shadow.g *= special.out_occlusion.g;
			special.out_specular_shadow.b *= special.out_occlusion.b;
		}
		else
		{
			special.out_specular_shadow.r = special.out_specular_shadow.r * special.out_occlusion.r * effect_spec + (1 - effect_spec) * special.out_specular_shadow.r;
			special.out_specular_shadow.g = special.out_specular_shadow.g * special.out_occlusion.g * effect_spec + (1 - effect_spec) * special.out_specular_shadow.g;
			special.out_specular_shadow.b = special.out_specular_shadow.b * special.out_occlusion.b * effect_spec + (1 - effect_spec) * special.out_specular_shadow.b;
		}
	}


													//Effect occlusion to diffuse
	float effect_diffuse = *mi_eval_scalar(&paras->effect_diffuse);
	if(sp->use_occlusion && (effect_diffuse > miEPS) && (alpha > miEPS))
	{
		if(effect_diffuse == 1)
		{
			special.out_color_occlusion.r = special.out_color.r * special.out_occlusion.r;
			special.out_color_occlusion.g = special.out_color.g * special.out_occlusion.g;
			special.out_color_occlusion.b = special.out_color.b * special.out_occlusion.b;
		}
		else
		{
			special.out_occlusion.r = special.out_occlusion.r * effect_diffuse + (1 - effect_diffuse);
			special.out_occlusion.g = special.out_occlusion.g * effect_diffuse + (1 - effect_diffuse);
			special.out_occlusion.b = special.out_occlusion.b * effect_diffuse + (1 - effect_diffuse);

			special.out_color_occlusion.r = special.out_color.r * special.out_occlusion.r;
			special.out_color_occlusion.g = special.out_color.g * special.out_occlusion.g;
			special.out_color_occlusion.b = special.out_color.b * special.out_occlusion.b;
		}
	}
	else
	{
		special.out_color_occlusion.r = special.out_color.r;
		special.out_color_occlusion.g = special.out_color.g;
		special.out_color_occlusion.b = special.out_color.b;
		special.out_occlusion.r = special.out_occlusion.g = special.out_occlusion.b = 1;
	}




	special.out_shading.r = special.out_diffuse_shadow.r * special.out_color_occlusion.r;			//Shading
	special.out_shading.g = special.out_diffuse_shadow.g * special.out_color_occlusion.g;
	special.out_shading.b = special.out_diffuse_shadow.b * special.out_color_occlusion.b;

//	special.out_color_occlusion.r = special.out_color.r * special.out_occlusion.r;										//Color_occlusion
//	special.out_color_occlusion.g = special.out_color.g * special.out_occlusion.g;
//	special.out_color_occlusion.b = special.out_color.b * special.out_occlusion.b;

	if(alpha < miEPS_INV)
	{
		if(mi_trace_transparent(result, state))
		{
			result->r *= inv_alpha;
			result->g *= inv_alpha;
			result->b *= inv_alpha;
		}
	}

	miScalar near = *mi_eval_scalar(&paras->near);
	miScalar far = *mi_eval_scalar(&paras->far);
	special.out_z.r = special.out_z.g = special.out_z.b = special.out_z.a = 0.0f;
	if(sp->use_z)																						//Z
	{
		miScalar depth;

		depth = (miScalar)state->dist;
		miScalar aver = 1.0f;

		if(sp->use_float_z == miFALSE)
			aver = far - near;

		switch(state->type)
		{
			case miRAY_EYE:
				if(depth < near)
					depth = near;
				if(depth > far)
					depth = far;
				depth = (depth - near)/aver;
				special.out_z.r = depth;
				break;
			case miRAY_TRANSPARENT:
				if(sp->opacity_z == miTRUE)
				{
					if(depth > far)
						depth = far;
					depth = depth / aver;
					special.out_z.r = depth;
				}
				break;
			case miRAY_REFLECT:
				if(sp->reflection_z == miTRUE)
				{
					if(depth > far)
						depth = far;
					special.out_z.g = depth / aver;
				}
				break;
			case miRAY_REFRACT:
				if(sp->refraction_z == miTRUE)
				{
					if(depth > far)
						depth = far;
					special.out_z.b = depth / aver;
				}
				break;
			default:
				break;
		}
	}


	if(alpha > miEPS)
	{
		//Result
		result->r += (special.out_diffuse_shadow.r * special.out_color_occlusion.r + special.out_incandescence.r + special.out_specular_shadow.r + special.out_indirect_illum.r + special.out_reflect.r + special.out_refract.r + special.out_amb_occlusion.r) * alpha;
		result->g += (special.out_diffuse_shadow.g * special.out_color_occlusion.g + special.out_incandescence.g + special.out_specular_shadow.g + special.out_indirect_illum.g + special.out_reflect.g + special.out_refract.g + special.out_amb_occlusion.g) * alpha;
		result->b += (special.out_diffuse_shadow.b * special.out_color_occlusion.b + special.out_incandescence.b + special.out_specular_shadow.b + special.out_indirect_illum.b + special.out_reflect.b + special.out_refract.b + special.out_amb_occlusion.b) * alpha;
		result->a += alpha;
		if(result->a > 1)
			result->a = 1;
		else
		{
			if(result->a < 0)
				result->a = 0;
		}

		if(sp->use_SpecialC1)											//SpecialC1
			special.out_specialC1 = *mi_eval_color(&paras->specialC1);
		if(sp->use_SpecialC2)											//SpecialC2
			special.out_specialC2 = *mi_eval_color(&paras->specialC2);
		if(sp->use_SpecialC3)											//SpecialC3
			special.out_specialC3 = *mi_eval_color(&paras->specialC3);
		if(sp->use_SpecialC4)											//SpecialC4
			special.out_specialC4 = *mi_eval_color(&paras->specialC4);
		if(sp->use_SpecialC5)											//SpecialC5
			special.out_specialC5 = *mi_eval_color(&paras->specialC5);
		if(sp->use_SpecialC6)											//SpecialC6
			special.out_specialC6 = *mi_eval_color(&paras->specialC6);
		if(sp->use_SpecialC7)											//SpecialC7
			special.out_specialC7 = *mi_eval_color(&paras->specialC7);
		if(sp->use_SpecialC8)											//SpecialC8
			special.out_specialC8 = *mi_eval_color(&paras->specialC8);
		if(sp->use_SpecialC9)											//SpecialC9
			special.out_specialC9 = *mi_eval_color(&paras->specialC9);
		if(sp->use_SpecialC10)											//SpecialC10
			special.out_specialC10 = *mi_eval_color(&paras->specialC10);
	}

		if(alpha < miEPS_INV && state->options->scanline != 'r')
		{
			special.out_shading.r *= alpha;
			special.out_shading.g *= alpha;
			special.out_shading.b *= alpha;

			special.out_color.r *= alpha;
			special.out_color.g *= alpha;
			special.out_color.b *= alpha;

			special.out_color_occlusion.r *= alpha;
			special.out_color_occlusion.g *= alpha;
			special.out_color_occlusion.b *= alpha;

			special.out_diffuse.r *= alpha;
			special.out_diffuse.g *= alpha;
			special.out_diffuse.b *= alpha;

			special.out_diffuse_shadow.r *= alpha;
			special.out_diffuse_shadow.g *= alpha;
			special.out_diffuse_shadow.b *= alpha;

			special.out_specular.r *= alpha;
			special.out_specular.g *= alpha;
			special.out_specular.b *= alpha;

			special.out_specular_shadow.r *= alpha;
			special.out_specular_shadow.g *= alpha;
			special.out_specular_shadow.b *= alpha;

			special.out_occlusion.r *= alpha;
			special.out_occlusion.g *= alpha;
			special.out_occlusion.b *= alpha;

			special.out_ambient.r *= alpha;
			special.out_ambient.g *= alpha;
			special.out_ambient.b *= alpha;

			special.out_amb_occlusion.r *= alpha;
			special.out_amb_occlusion.g *= alpha;
			special.out_amb_occlusion.b *= alpha;

			special.out_indirect_illum.r *= alpha;
			special.out_indirect_illum.g *= alpha;
			special.out_indirect_illum.b *= alpha;

			special.out_reflect.r *= alpha;
			special.out_reflect.g *= alpha;
			special.out_reflect.b *= alpha;

			special.out_refract.r *= alpha;
			special.out_refract.g *= alpha;
			special.out_refract.b *= alpha;

			special.out_incandescence.r *= alpha;
			special.out_incandescence.g *= alpha;
			special.out_incandescence.b *= alpha;

			special.out_z.g *= alpha;
			special.out_z.b *= alpha;

			special.out_bent_normal.r *= alpha;
			special.out_bent_normal.g *= alpha;
			special.out_bent_normal.b *= alpha;

			special.out_specialC1.r *= alpha;
			special.out_specialC1.g *= alpha;
			special.out_specialC1.b *= alpha;

			special.out_specialC2.r *= alpha;
			special.out_specialC2.g *= alpha;
			special.out_specialC2.b *= alpha;

			special.out_specialC3.r *= alpha;
			special.out_specialC3.g *= alpha;
			special.out_specialC3.b *= alpha;

			special.out_specialC4.r *= alpha;
			special.out_specialC4.g *= alpha;
			special.out_specialC4.b *= alpha;

			special.out_specialC5.r *= alpha;
			special.out_specialC5.g *= alpha;
			special.out_specialC5.b *= alpha;

			special.out_specialC6.r *= alpha;
			special.out_specialC6.g *= alpha;
			special.out_specialC6.b *= alpha;

			special.out_specialC7.r *= alpha;
			special.out_specialC7.g *= alpha;
			special.out_specialC7.b *= alpha;

			special.out_specialC8.r *= alpha;
			special.out_specialC8.g *= alpha;
			special.out_specialC8.b *= alpha;

			special.out_specialC9.r *= alpha;
			special.out_specialC9.g *= alpha;
			special.out_specialC9.b *= alpha;

			special.out_specialC10.r *= alpha;
			special.out_specialC10.g *= alpha;
			special.out_specialC10.b *= alpha;
		}
//	}

		if((state->type == miRAY_EYE) || (state->type == miRAY_TRANSPARENT))			//Writing buffers
		{
			mi::shader::Access_fb fbc(state->camera->buffertag);
			size_t index = -1;
			miColor temp_color;


			if(sp->shading_pass)
			{
				if(sp->shading_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_shading", index)) 
						sp->shading_buf = (int)index;
					else
						sp->shading_pass = miFALSE;				
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->shading_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_shading.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_shading.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_shading.b;
				mi_fb_put(state, sp->shading_buf, &temp_color);
			}
			if(sp->color_pass)
			{
				if(sp->color_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_color", index)) 
						sp->color_buf = (int)index;
					else 
						sp->color_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->color_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_color_occlusion.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_color_occlusion.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_color_occlusion.b;
				mi_fb_put(state, sp->color_buf, &temp_color);
			}
			if(sp->diffuse_pass)
			{
				if(sp->diffuse_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_diffuse", index)) 
						sp->diffuse_buf = (int)index;
					else
						sp->diffuse_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->diffuse_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_diffuse.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_diffuse.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_diffuse.b;
				mi_fb_put(state, sp->diffuse_buf, &temp_color);
			}
			if(sp->diffuse_shadow_pass)
			{
				if(sp->diffuse_shadow_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_shadow", index)) 
						sp->diffuse_shadow_buf  = (int)index;
					else
						sp->diffuse_shadow_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->diffuse_shadow_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_diffuse_shadow.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_diffuse_shadow.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_diffuse_shadow.b;
				mi_fb_put(state, sp->diffuse_shadow_buf, &temp_color);
			}
			if(sp->specular_pass)
			{
				if(sp->specular_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specular", index)) 
						sp->specular_buf = (int)index;
					else
						sp->specular_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specular_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specular.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specular.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specular.b;
				mi_fb_put(state, sp->specular_buf, &temp_color);
			}
			if(sp->specular_shadow_pass)
			{
				if(sp->specular_shadow_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_spec_shd", index)) 
						sp->specular_shadow_buf = (int)index;
					else
						sp->specular_shadow_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specular_shadow_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specular_shadow.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specular_shadow.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specular_shadow.b;
				mi_fb_put(state, sp->specular_shadow_buf, &temp_color);
			}
			if(sp->occlusion_pass)
			{
				if(sp->occlusion_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_diff_occl", index)) 
						sp->occlusion_buf = (int)index;
					else
						sp->occlusion_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->occlusion_buf, &temp_color);
				if(alpha > miEPS)
				{
					temp_color.r = temp_color.r * inv_alpha + special.out_occlusion.r;
					temp_color.g = temp_color.g * inv_alpha + special.out_occlusion.g;
					temp_color.b = temp_color.b * inv_alpha + special.out_occlusion.b;
				}
				mi_fb_put(state, sp->occlusion_buf, &temp_color);
			}
			if(sp->ambient_pass)
			{
				if(sp->ambient_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_ambient", index)) 
						sp->ambient_buf = (int)index;
					else
						sp->ambient_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->ambient_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_ambient.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_ambient.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_ambient.b;
				mi_fb_put(state, sp->ambient_buf, &temp_color);
			}
			if(sp->amb_occlusion_pass)
			{
				if(sp->amb_occlusion_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_amb_occlus", index))
						sp->amb_occlusion_buf = (int)index;
					else
						sp->amb_occlusion_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->amb_occlusion_buf, &temp_color);
				if(alpha > miEPS)
				{
					temp_color.r = temp_color.r * inv_alpha + special.out_amb_occlusion.r;
					temp_color.g = temp_color.g * inv_alpha + special.out_amb_occlusion.g;
					temp_color.b = temp_color.b * inv_alpha + special.out_amb_occlusion.b;
				}
				mi_fb_put(state, sp->amb_occlusion_buf, &temp_color);
			}
			if(sp->indirect_illum_pass)
			{
				if(sp->indirect_illum_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_indirect", index)) 
						sp->indirect_illum_buf = (int)index;
					else
						sp->indirect_illum_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->indirect_illum_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_indirect_illum.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_indirect_illum.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_indirect_illum.b;
				mi_fb_put(state, sp->indirect_illum_buf, &temp_color);
			}
			if(sp->reflect_pass)
			{
				if(sp->reflect_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_reflect", index)) 
						sp->reflect_buf = (int)index;
					else
						sp->reflect_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->reflect_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_reflect.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_reflect.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_reflect.b;
				mi_fb_put(state, sp->reflect_buf, &temp_color);
			}
			if(sp->refract_pass)
			{
				if(sp->refract_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_refract", index)) 
						sp->refract_buf = (int)index;
					else
						sp->refract_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->refract_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_refract.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_refract.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_refract.b;
				mi_fb_put(state, sp->refract_buf, &temp_color);
			}
			if(sp->incandescence_pass)
			{
				if(sp->incandescence_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_incand", index)) 
						sp->incandescence_buf = (int)index;
					else
						sp->incandescence_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->incandescence_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_incandescence.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_incandescence.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_incandescence.b;
				mi_fb_put(state, sp->incandescence_buf, &temp_color);
			}
			if(sp->bent_normal_pass)
			{
				if(sp->bent_normal_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_bent", index)) 
						sp->bent_normal_buf = (int)index;
					else
						sp->bent_normal_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->bent_normal_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_bent_normal.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_bent_normal.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_bent_normal.b;
				mi_fb_put(state, sp->bent_normal_buf, &temp_color);
			}
			if(sp->specialC1_pass)
			{
				if(sp->specialC1_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC1", index)) 
						sp->specialC1_buf = (int)index;
					else
						sp->specialC1_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC1_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC1.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC1.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC1.b;
				mi_fb_put(state, sp->specialC1_buf, &temp_color);
			}
			if(sp->specialC2_pass)
			{
				if(sp->specialC2_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC2", index)) 
						sp->specialC2_buf = (int)index;
					else
						sp->specialC2_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC2_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC2.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC2.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC2.b;
				mi_fb_put(state, sp->specialC2_buf, &temp_color);
			}
			if(sp->specialC3_pass)
			{
				if(sp->specialC3_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC3", index))
						sp->specialC3_buf = (int)index;
					else
						sp->specialC3_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC3_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC3.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC3.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC3.b;
				mi_fb_put(state, sp->specialC3_buf, &temp_color);
			}
			if(sp->specialC4_pass)
			{
				if(sp->specialC4_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC4", index)) 
						sp->specialC4_buf = (int)index;
					else
						sp->specialC4_pass = miFALSE;

				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC4_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC4.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC4.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC4.b;
				mi_fb_put(state, sp->specialC4_buf, &temp_color);
			}
			if(sp->specialC5_pass)
			{
				if(sp->specialC5_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC5", index))
						sp->specialC5_buf = (int)index;
					else
						sp->specialC5_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC5_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC5.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC5.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC5.b;
				mi_fb_put(state, sp->specialC5_buf, &temp_color);
			}
			if(sp->specialC6_pass)
			{
				if(sp->specialC6_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC6", index)) 
						sp->specialC6_buf = (int)index;
					else
						sp->specialC6_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC6_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC6.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC6.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC6.b;
				mi_fb_put(state, sp->specialC6_buf, &temp_color);
			}
			if(sp->specialC7_pass)
			{
				if(sp->specialC7_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC7", index)) 
						sp->specialC7_buf = (int)index;
					else
						sp->specialC7_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC7_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC7.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC7.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC7.b;
				mi_fb_put(state, sp->specialC7_buf, &temp_color);
			}
			if(sp->specialC8_pass)
			{
				if(sp->specialC8_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC8", index)) 
						sp->specialC8_buf = (int)index;
					else
						sp->specialC8_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC8_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC8.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC8.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC8.b;
				mi_fb_put(state, sp->specialC8_buf, &temp_color);
			}
			if(sp->specialC9_pass)
			{
				if(sp->specialC9_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC9", index)) 
						sp->specialC9_buf = (int)index;
					else
						sp->specialC9_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC9_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC9.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC9.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC9.b;
				mi_fb_put(state, sp->specialC9_buf, &temp_color);
			}
			if(sp->specialC10_pass)
			{
				if(sp->specialC10_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC10", index))
						sp->specialC10_buf = (int)index;
					else
						sp->specialC10_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC10_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC10.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC10.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC10.b;
				mi_fb_put(state, sp->specialC10_buf, &temp_color);
			}
			if(sp->z_pass)
			{
				if(sp->z_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_z", index)) 
						sp->z_buf = (int)index;
					else
						sp->z_pass = miFALSE;

				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->z_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_z.r;
				temp_color.g = temp_color.g + special.out_z.g;
				temp_color.b = temp_color.b + special.out_z.b;

				if(sp->invert_z == miTRUE)
				{
					if(state->type == miRAY_EYE)
					{
						if(sp->use_float_z == miFALSE)
						{
							temp_color.r = 1 -temp_color.r;
							temp_color.g = 1 -temp_color.g;
							temp_color.b = 1 -temp_color.b;
						}
						else
						{
							temp_color.r = (far - near) -temp_color.r;
							temp_color.g = (far - near) -temp_color.g;
							temp_color.b = (far - near) -temp_color.b;
						}
					}
				}
				temp_color.a += special.out_z.a;
				mi_fb_put(state, sp->z_buf, &temp_color);
			}	
		}
	

	if(1)	//check for NaN value
	{
		#ifdef WIN32
			if(_isnan(result->r) || !_finite(result->r))
			{
				mi_warning("p_MegaTK shader return NaN or infinite R value. Replaced with zero");
				result->r = 0;
			}
			if(_isnan(result->g) || !_finite(result->g))
			{
				mi_warning("p_MegaTK shader return NaN or infinite G value. Replaced with zero");
				result->g = 0;
			}
			if(_isnan(result->b) || !_finite(result->b))
			{
				mi_warning("p_MegaTK shader return NaN or infinite B value. Replaced with zero");
				result->b = 0;
			}
			if(_isnan(result->a) || !_finite(result->a))
			{
				mi_warning("p_MegaTK shader return NaN or infinite A value. Replaced with zero");
				result->a = 0;
			}
		#else	//linux or mac
			if(!check_nan (result->r))
			{
				mi_warning("p_MegaTK shader return NaN or infinite R value. Replaced with zero");
				result->r = 0;
			}
			if(!check_nan (result->g))
			{
				mi_warning("p_MegaTK shader return NaN or infinite G value. Replaced with zero");
				result->g = 0;
			}
			if(!check_nan (result->b))
			{
				mi_warning("p_MegaTK shader return NaN or infinite B value. Replaced with zero");
				result->b = 0;
			}
			if(!check_nan (result->a))
			{
				mi_warning("p_MegaTK shader return NaN or infinite A value. Replaced with zero");
				result->a = 0;
			}
		#endif

	}
	//Set opacity for rasterizer
	if(state->options->scanline == 'r')
	{
		miColor rasterizer_opacity;
		rasterizer_opacity.r = rasterizer_opacity.g = rasterizer_opacity.b = rasterizer_opacity.a = alpha;
		mi_opacity_set(state, &rasterizer_opacity);
	}

// mi_warning("after sample: numSamples=%d", numSamples);
// mi_warning("after all samples: numSamples=%d", numSamples);

//	`REAL_VALUE_ISNAN (X)'

//	if(_isnan(result->r))
//		mi_warning("awawwawa");

	return(miTRUE);
}
}
