#pragma once

#ifndef p_deriv_h
#define p_deriv_h

#include <math.h>
#include "shader.h"

inline
void get_vertices( const miState* const state, miVector v[3] )
{
   //
   // First, obtain the points of the triangle
   //
   const miVector* p[3];
   mi_tri_vectors(const_cast<miState*>(state), 'p', 0, &p[0], &p[1], &p[2]);

   // Those points come in object space, transform to internal...
   mi_point_from_object( const_cast<miState*>(state), &v[0], p[0] );
   mi_point_from_object( const_cast<miState*>(state), &v[1], p[1] );
   mi_point_from_object( const_cast<miState*>(state), &v[2], p[2] );
}

static
miBoolean dPdst(miVector& dPds, miVector& dPdt,	const miState* const state, const int idx = 0)
{   
   int num;
   mi_query( miQ_NUM_TEXTURES, const_cast<miState*>( state ), miNULLTAG, &num );
   if ( idx >= num )
   {
		mi_warning("Object does not have ST coordinates!");
		dPds.x = dPds.y = dPds.z = dPdt.x = dPdt.y = dPdt.z = 0.0f;
		return miFALSE;
   }
   
  //
  // First, obtain the texture coordinates of the the three
  // vertices of the triangle, according to user parameters 
  // (projection type, etc).
  //
  const miVector* t[3];
  mi_tri_vectors(const_cast< miState* >(state), 't', idx, &t[0], &t[1], &t[2]);

  miScalar u1 = t[1]->x - t[0]->x;
  miScalar u2 = t[2]->x - t[0]->x;
  miScalar v1 = t[1]->y - t[0]->y;
  miScalar v2 = t[2]->y - t[0]->y;
  
  miScalar det = u1 * v2 - v1 * u2;
  if (det == 0.0f)
  {
     // degenerate
     dPds.x = dPdt.y = 1;
     dPds.y = dPds.z = dPdt.x = dPdt.z = 0;
     return miFALSE;
  }
  
  //
  // First, obtain the points of the triangle
  // 
  miVector v[3];
  get_vertices( state, v );
  
  miScalar x1 = v[1].x - v[0].x;
  miScalar x2 = v[2].x - v[0].x;
  miScalar y1 = v[1].y - v[0].y;
  miScalar y2 = v[2].y - v[0].y;
  miScalar z1 = v[1].z - v[0].z;
  miScalar z2 = v[2].z - v[0].z;
  
  miScalar invdet = 1.0f / det;
  dPds.x = (x1 * v2 - v1 * x2) * invdet;
  dPds.y = (y1 * v2 - v1 * y2) * invdet;
  dPds.z = (z1 * v2 - v1 * z2) * invdet;
  dPdt.x = (u1 * x2 - x1 * u2) * invdet;
  dPdt.y = (u1 * y2 - y1 * u2) * invdet;
  dPdt.z = (u1 * z2 - z1 * u2) * invdet;
  return miTRUE;
}

static 
miBoolean BumpUV(miVector& dPdu, miVector& dPdv, const miState* const state, const int idx = 0)
{
	// Check to see if bump derivatives available.  If so, use that instead, as
	// they are more accurate.
	int num = 0;
	mi_query( miQ_NUM_BUMPS, const_cast<miState*>(state), miNULLTAG, &num);

	if ( num > idx && state->bump_x_list != NULL )
	{
		dPdu = state->bump_x_list[idx];
		dPdv = state->bump_y_list[idx];
		return miTRUE;
	}

	if ( ! dPdst( dPdu, dPdv, state, idx ) ) return miFALSE;
   
	// Make sure tangents form an orthogonal basis with normal.
	// See Ken Turkowski '93
	// http://www.worldserver.com/turk/computergraphics/DifferentialMappings.pdf
	miVector C;
	mi_vector_prod(&C, &state->normal, &dPdu );
	miScalar t2 = dPdu.x * dPdu.x + dPdu.y * dPdu.y + dPdu.z * dPdu.z;
	miScalar NxT2 = C.x * C.x + C.y * C.y + C.z * C.z;
	mi_vector_prod(&dPdu, &C, &state->normal);
	miScalar t2NxT2 = ( t2 / NxT2 );
//	dPdu *= ( t2 / NxT2 );
	dPdu.x *= t2NxT2;
	dPdu.y *= t2NxT2;
	dPdu.z *= t2NxT2;
   
	mi_vector_prod(&C, &state->normal, &dPdv );
	t2 = dPdv.x * dPdv.x + dPdv.y * dPdv.y + dPdv.z * dPdv.z;
	NxT2 = C.x * C.x + C.y * C.y + C.z * C.z;
	mi_vector_prod(&dPdv, &C, &state->normal);
	t2NxT2 = ( t2 / NxT2 );
//	dPdv *= ( t2 / NxT2 );
	dPdv.x *= t2NxT2;
	dPdv.y *= t2NxT2;
	dPdv.z *= t2NxT2;

   return miTRUE;
}

inline void p_UV_orientation(miVector *u, miVector *v, miState *state)
{
	float d;
//	#ifdef MAYA
//		*u = state->bump_x_list[0];
//	#else
//		miVector dummy;
//		BumpUV( *u, dummy, state, 0);
//		mi_vector_normalize(u);
//	#endif

	miVector dummy;
	BumpUV( *u, dummy, state, 0);
	mi_vector_normalize(u);

	d  = mi_vector_dot(u, &state->normal);
	u->x -= d * state->normal.x;
	u->y -= d * state->normal.y;
	u->z -= d * state->normal.z;
	miASSERT(mi_vector_dot(u, &state->normal) < miEPS);
	mi_vector_normalize(u);

	// Set v to be perpendicular to u (in the tangent plane) 
	mi_vector_prod(v, &state->normal, u);
	miASSERT(fabs(mi_vector_norm(u) - 1) < miEPS);
	miASSERT(fabs(mi_vector_norm(v) - 1) < miEPS);
	miASSERT(fabs(mi_vector_dot(u, v)) < miEPS);
}

#endif  // p_deriv_h
