//  Copyright (c) 2005, Pavel Ledin and by Daniel "thebigMuh" Rind
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin and Daniel Rind nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//Shader based on "Dirtmap" shader by Daniel "thebigMuh" Rind

#include <math.h>
#include "shader.h"

typedef struct
{
	miScalar	coefficient;
	miColor		occlusion;
	miScalar	spread;
	int			samples;
	miBoolean	use_normal;
	miVector	direction;
}p_environment_t;


static inline miVector
p_sphere_sample(miGeoScalar sample[2])
{
	// Compute a unit direction vector from 2 samples on a spherical hull:
	miVector	smp;
	double		t, st, phirad;

	t = 2.0 * sample[0] - 1.0;			// map from [0, 1) to [-1, 1)
	st = sqrt(1.0 - t*t);

	phirad = sample[1] * 2.0 * M_PI;	// map from [0, 1) to [0, 2pi)

	smp.x = (miScalar)(st * cos(phirad));
	smp.y = (miScalar)(st * sin(phirad));
	smp.z = (miScalar)t;

	return smp;
}

extern "C"
{
DLLEXPORT int p_environment_version(void) {return(2);}

DLLEXPORT miBoolean p_environment(miColor *result, miState *state,	p_environment_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	double		sample[2];		// 2d sample, will be filled by mi_sample
	miInteger	sampleCnt;		// sample counter, needed for mi_sample
	miVector	sampleDir, n;	// unit direction vector ray will be shot at, and the current normal n
	miColor		hitSum;
	miColor		curEnvironment;
	miScalar	spread = *mi_eval_scalar(&paras->spread);
	miScalar	coeff = *mi_eval_scalar(&paras->coefficient);

	miUint rayCount = *mi_eval_integer(&paras->samples);
	miColor *occlusion = mi_eval_color(&paras->occlusion);

	if (rayCount <= 0)
	{
		result->r = result->g = result->b = 0;
		return miTRUE;
	}

	sampleCnt = 0;

	hitSum.r = hitSum.g = hitSum.b = 0;

	if(*mi_eval_boolean(&paras->use_normal))
		n = state->normal;
	else
		n = *mi_eval_vector(&paras->direction);

	if(spread == 0)
	{
		mi_trace_environment(&hitSum, state, &n);

		result->r = hitSum.r * coeff * occlusion->r;
		result->g = hitSum.g * coeff * occlusion->g;
		result->b = hitSum.b * coeff * occlusion->b;
	}
	else
	{
		mi_vector_mul(&n, 1.0f - (spread / 180.0f));
		// get a 2-d sample from mi's QMC sampler
		while(mi_sample(sample, &sampleCnt, state, 2, &rayCount))
		{
			// map the sample to a sphere
			sampleDir = p_sphere_sample(sample);

			// premultiply this
			mi_vector_mul(&sampleDir, spread / 180.0f);

			// add it together
			mi_vector_add(&sampleDir, &n, &sampleDir);

			mi_trace_environment(&curEnvironment, state, &sampleDir);
			hitSum.r += curEnvironment.r;
			hitSum.g += curEnvironment.g;
			hitSum.b += curEnvironment.b;
		}
		result->r = hitSum.r / sampleCnt * coeff * occlusion->r;
		result->g = hitSum.g / sampleCnt * coeff * occlusion->g;
		result->b = hitSum.b / sampleCnt * coeff * occlusion->b;
	}

	return(miTRUE);
}
}
