//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
   miColor input;
   miColor pointBlack;
   miColor pointWhite;
   miColor saturation;
   int saturation_mode;
   miScalar gain;
   miScalar gamma;
   miColor mult;
   miColor offset;
   miBoolean clampBlack;
   miBoolean clampWhite;
   miBoolean invert;
}p_grade_color_t;

extern "C"
{
DLLEXPORT int p_grade_color_version(void) {return(2);}

DLLEXPORT miBoolean p_grade_color(miColor *result, miState *state, p_grade_color_t *paras)
{
	//eval...
	*result =				*mi_eval_color(&paras->input);
	
	miColor *pointBlack =	mi_eval_color(&paras->pointBlack);
	miColor *pointWhite =	mi_eval_color(&paras->pointWhite);
	
	miColor *saturation =	mi_eval_color(&paras->saturation);
	int mode =				*mi_eval_integer(&paras->saturation_mode);

	miScalar gain =			*mi_eval_scalar(&paras->gain);
	miScalar gamma =		*mi_eval_scalar(&paras->gamma);

	miColor *mult =			mi_eval_color(&paras->mult);
	miColor *offset =		mi_eval_color(&paras->offset);

	miBoolean clampBlack =	*mi_eval_boolean(&paras->clampBlack);
	miBoolean clampWhite =	*mi_eval_boolean(&paras->clampWhite);

	miBoolean invert =		*mi_eval_boolean(&paras->invert);


	//grade...
	result->r = (result->r - pointBlack->r) / (pointWhite->r - pointBlack->r);
	result->g = (result->g - pointBlack->g) / (pointWhite->g - pointBlack->g);
	result->b = (result->b - pointBlack->b) / (pointWhite->b - pointBlack->b);
      
	if(saturation->r != 1.0f || saturation->g != 1.0f || saturation->b != 1.0f)			//saturation
	{
		miScalar gray;
		if(mode)		//luminosity
			gray = result->r*0.299f + result->g*0.587f + result->b*0.114f;
//			gray = result->r*0.2125f + result->g*0.7154f* + result->b*.0721f*;
		else			//average
			gray = (result->r + result->g + result->b)/3;


		if(saturation->r != 1.f)				//saturation R
		{
			if(saturation->r == 0)
				result->r = gray;
			else
				result->r = gray*(1-saturation->r) + result->r*saturation->r;
		}
		if(saturation->g != 1.f)				//saturation G
		{
			if(saturation->g == 0)
				result->g = gray;
			else
				result->g = gray*(1-saturation->g) + result->g*saturation->g;
		}
		if(saturation->b != 1.f)				//saturation B
		{
			if(saturation->b == 0)
				result->b = gray;
			else
				result->b = gray*(1-saturation->b) + result->b*saturation->b;
		}
	}

	if(gain != 1.0f)				//gain
	{
		result->r *= gain; 
		result->g *= gain; 
		result->b *= gain;
	}

	if(gamma != 1.0f)				//gamma
	{
		if(result->r > 0) 
			result->r = pow(result->r, 1/gamma);
		if(result->g > 0) 
			result->g = pow(result->g, 1/gamma);
		if(result->b > 0) 
			result->b = pow(result->b, 1/gamma);
	}
									
	result->r *= mult->r;			//mult
	result->g *= mult->g; 
	result->b *= mult->b;

	result->r += offset->r;			//offset
	result->g += offset->g; 
	result->b += offset->b;

    if(invert)						//invert
	{ 
		result->r = 1 - result->r; 
		result->g = 1 - result->g; 
		result->b = 1 - result->b;
	}

	if(clampBlack)					//clamp black
	{
		if(result->r < 0) 
			result->r = 0;
		if(result->g < 0) 
			result->g = 0;
		if(result->b < 0) 
			result->b = 0;
	}
	if(clampWhite)					//clamp white
	{
		if(result->r > 1) 
			result->r = 1.0f;
		if(result->g > 1) 
			result->g = 1.0f;
		if(result->b > 1) 
			result->b = 1.0f;
	}
	return(miTRUE);
}
}
