//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miTag buffer_a;
	miTag buffer_b;
	int operation;
}p_math_output_t;

extern "C"
{
DLLEXPORT int p_math_output_version(void) {return(2);}

DLLEXPORT miBoolean p_math_output(void *result, register miState *state, p_math_output_t *paras)
{
	int x, y;
	miColor color_a;
	miColor color_b;
	miColor color;

	miTag a_tag = *mi_eval_tag(&paras->buffer_a);				//buffer_a
	if (a_tag == miNULLTAG)
		return miFALSE;
	const char *a_name = (char*)mi_db_access(a_tag);
	mi_db_unpin(a_tag);

	miTag b_tag = *mi_eval_tag(&paras->buffer_b);				//buffer_a
	if (b_tag == miNULLTAG)
		return miFALSE;
	const char *b_name = (char*)mi_db_access(b_tag);
	mi_db_unpin(b_tag);

	mi::shader::Access_fb fbc(state->camera->buffertag);

	size_t a_index = -1;
	size_t b_index = -1;

	fbc->get_index(a_name, a_index);
	fbc->get_index(b_name, b_index);

	int a;
	int b;
	if(a_index != -1)
		a = (int)a_index;
	else
	{
		mi_warning("p_math_output: wrong buffer number for buffer %s", a_name);
		return(miFALSE);
	}

	if(b_index != -1)
		b = (int)b_index;
	else 
	{
		mi_warning("p_math_output: wrong buffer number for buffer %s", b_name);
		return(miFALSE);
	}

	int operat = *mi_eval_integer(&paras->operation);

	miImg_image *fb_color_a = mi_output_image_open(state, a + miRC_IMAGE_USER);
	miImg_image *fb_color_b = mi_output_image_open(state, b + miRC_IMAGE_USER);

	switch(operat)
	{
		case 0:		//multiply
			mi_progress("p_math_output: %s * %s", a_name, b_name);
			break;
		case 1:		//divide					
			mi_progress("p_math_output: %s / %s", a_name, b_name);
			break;
		case 2:		//plus
			mi_progress("p_math_output: %s + %s", a_name, b_name);
			break;
		case 3:		//minus
			mi_progress("p_math_output: %s - %s", a_name, b_name);
			break;
		default:
			break;
	}

	for (y=0; y < state->camera->y_resolution; y++)
	{
		if(mi_par_aborted())
			break;
		for (x=0; x<state->camera->x_resolution; x++)
		{
			mi_img_get_color(fb_color_a, &color_a, x, y);
			mi_img_get_color(fb_color_b, &color_b, x, y);
			switch(operat)
			{
				case 0:		//multiply
					color.r = color_a.r * color_b.r;
					color.g = color_a.g * color_b.g;
					color.b = color_a.b * color_b.b;
					break;
				case 1:		//divide
					if(color_b.r <= miSCALAR_EPSILON || color_a.r <= miSCALAR_EPSILON)
						color.r = 0;
					else
						color.r = color_a.r / color_b.r;

					if(color_b.g <= miSCALAR_EPSILON || color_a.g <= miSCALAR_EPSILON)
						color.g = 0;
					else
						color.g = color_a.g / color_b.g;

					if(color_b.b <= miSCALAR_EPSILON || color_a.b <= miSCALAR_EPSILON)
						color.b = 0;
					else
						color.b = color_a.b / color_b.b;
					break;
				case 2:		//plus
					color.r = color_a.r + color_b.r;
					color.g = color_a.g + color_b.g;
					color.b = color_a.b + color_b.b;
					break;
				case 3:		//minus
					color.r = color_a.r - color_b.r;
					color.g = color_a.g - color_b.g;
					color.b = color_a.b - color_b.b;
					break;
				default:
					break;
			}
			mi_img_put_color(fb_color_a, &color, x, y);
		}
	}
	mi_output_image_close(state, a + miRC_IMAGE_USER);
	mi_output_image_close(state, b + miRC_IMAGE_USER);
	return(miTRUE);
}
}
