//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miScalar	tnear;
	miScalar	tfar;
	miScalar	opacity;
	miScalar	reflection;
	miScalar	shiny_reflect;
	miScalar	refraction;
	miScalar	shiny_refract;
	miScalar	ior;
	miBoolean	use_float;		// 0-1 or float
	miBoolean	invert;
	miBoolean	split_color_result;
	miBoolean	alpha_mult;
}depth_t;

#define miEPS 0.0001

extern "C"
{
DLLEXPORT int p_z_version(void) {return(2);}

DLLEXPORT miBoolean p_z(miColor *result, miState *state, depth_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miScalar *near;
	miScalar *far;
	miBoolean *pfloat;
	miScalar alpha;
	miScalar depth;
	miBoolean *pinvert;

	near = mi_eval_scalar(&paras->tnear);
	far = mi_eval_scalar(&paras->tfar);
	pfloat = mi_eval_boolean(&paras->use_float);
	pinvert = mi_eval_boolean(&paras->invert);
	alpha = *mi_eval_scalar(&paras->opacity);
	miBoolean split = *mi_eval_boolean(&paras->split_color_result);
	miScalar reflect = *mi_eval_scalar(&paras->reflection);
	miScalar refract = *mi_eval_scalar(&paras->refraction);

    depth = (miScalar)state->dist;

	miScalar aver = 1;

	if(*pfloat == miFALSE)
		aver = *far - *near;

	if(state->type == miRAY_EYE)
	{
		if(depth < *near)
			depth = *near;
		if(depth > *far)
			depth = *far;
		depth = (depth - *near)/aver;
	}
	else
	{
		if(depth > *far)
			depth = *far;
		depth = depth / aver;
	}

	result->a = 0;
	miScalar reflcol = 0;
	miScalar refrcol = 0;

	if(alpha < 1)					//Transp
	{
		result->a = alpha;
		miColor transp_res;
		mi_trace_transparent(&transp_res, state);
		if(split == miTRUE)
		{
			depth += transp_res.r * (1-alpha);
			reflcol += transp_res.g * (1-alpha);
			refrcol += transp_res.b * (1-alpha);
		}
		else
			depth += transp_res.r * (1-alpha);
		result->a += transp_res.a;
	}
	else
		result->a = 1;

//	miVector rdir;

	if(alpha > miEPS)
	{
		if(reflect > miEPS)					//Reflect
		{
			miVector rdir;
			miColor rcol;
			if(*mi_eval_scalar(&paras->shiny_reflect) > 0)
			{
				mi_reflection_dir_glossy(&rdir, state, *mi_eval_scalar(&paras->shiny_reflect));
				if(mi_trace_reflection(&rcol, state, &rdir))
				{
					if(split == miTRUE)
						reflcol = (rcol.r + rcol.g + rcol.b) * reflect * alpha;
					else
						reflcol += rcol.r * reflect * alpha;
				}
			}
			else
			{
				mi_reflection_dir(&rdir, state);
				if(mi_trace_reflection(&rcol, state, &rdir))
				{
					if(split == miTRUE)
						reflcol = (rcol.r + rcol.g + rcol.b) * reflect * alpha;
					else
						reflcol += rcol.r * reflect * alpha;
				}
			}
		}
		if(refract > miEPS)					//Refract
		{
	//		mi_warning("%s", "aaaa");
			miVector rdir;
			miColor rcol;
			if(*mi_eval_scalar(&paras->shiny_refract) > 0)
			{
				mi_transmission_dir_glossy(&rdir, state, 1.0, *mi_eval_scalar(&paras->ior), *mi_eval_scalar(&paras->shiny_refract));
				if(mi_trace_refraction(&rcol, state, &rdir))
				{
					if(split == miTRUE)
						refrcol = (rcol.r + rcol.g + rcol.b) * refract * alpha;
					else
						refrcol = rcol.r * refract * alpha;
				}
			}
			else
			{
				mi_refraction_dir(&rdir, state, 1.0, *mi_eval_scalar(&paras->ior));
				if(mi_trace_refraction(&rcol, state, &rdir))
				{
					if(split == miTRUE)
						refrcol = (rcol.r + rcol.g + rcol.b) * refract * alpha;
					else
						refrcol = rcol.r * refract * alpha;
				}
			}
		}
	}

	if(split == miTRUE)
	{
		result->r = depth;					//primary z
		result->g = reflcol;				//reflection z
		result->b = refrcol;				//refraction z
	}
	else
		result->r = result->g = result->b = depth + reflcol + refrcol;



	if(*pinvert == miTRUE)					//Invert
	{
		if(state->type == miRAY_EYE)
		{
			if(*pfloat == miFALSE)
			{
				result->r = 1 -result->r;
				result->g = 1 -result->g;
				result->b = 1 -result->b;
			}
			else
			{
				result->r = (*far - *near) -result->r;
				result->g = (*far - *near) -result->g;
				result->b = (*far - *near) -result->b;
			}
			if(result->a < 1 && *mi_eval_boolean(&paras->alpha_mult))
			{
				result->r *= result->a;
				result->g *= result->a;
				result->b *= result->a;
			}
		}
	}
	else
	{
		if(state->type == miRAY_EYE && result->a < 1 && *mi_eval_boolean(&paras->alpha_mult))
		{
			result->r *= result->a;
			result->g *= result->a;
			result->b *= result->a;
		}
	}

	return(miTRUE);
}
}
